from __future__ import annotations

import copy
from abc import ABC, abstractmethod
from typing import Any, Sequence, Optional, TYPE_CHECKING, TypeVar
from uuid import UUID

import pandas as pd

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixin, DataObjectMixinDictType
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.FileOperations.File import File
from ResSimpy.Nexus.DataModels.Network.NexusConstraint import NexusConstraint
from ResSimpy.Time.ISODateTime import ISODateTime
from ResSimpy.Utils.obj_to_table_string import get_column_headers_required

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import Network

T = TypeVar('T', bound=DataObjectMixin)


class NetworkOperationsMixIn(ABC):
    """A mixin for network collections that forces the implementation of the same methods across all classes."""
    def __init__(self, parent_network: Network) -> None:
        """Initialises the NetworkOperationsMixIn class.

        Args:
            parent_network (Network): The parent network that the object is a part of.
        """
        self.__parent_network = parent_network

    @abstractmethod
    def get_all(self) -> Sequence[Any]:
        """Returns list of network operations."""
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_by_name(self, name: str) -> Optional[Any]:
        """Returns network object with provided name.

        Args:
            name(str): Name of network object.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_df(self) -> pd.DataFrame:
        """Creates dataframe representing network operations in a file."""
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_overview(self) -> str:
        """Returns an overview of the network operations."""
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def load(self, file: File, start_date: str, default_units: UnitSystem) -> None:
        """Loads file path using the default units.

        Args:
            file(File): Loads file path
            start_date(str): Starting date of the run.
            default_units(UnitSystem): Units used if not specified by the file.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def _add_to_memory(self, additional_objs: Optional[list[Any]]) -> None:
        """Adds an instance of objects to the current list.

        Args:
            additional_objs(Optional[list[Any]]): New objects to be added.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def remove(self, obj_to_remove: UUID | DataObjectMixinDictType) -> None:
        """Removes the UUID from the network.

        Args:
            obj_to_remove(UUID | DataObjectMixinDictType): UUID of the network to be removed
            from the dictionary.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def add(self, obj_to_add: DataObjectMixinDictType) -> None:
        """Adding a connection to a network taking a dictionary with properties for the new connection.

        Args:
            obj_to_add(DataObjectMixinDictType): dictionary taking all the new properties for the network.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def modify(self, obj_to_modify: DataObjectMixinDictType,
               new_properties: DataObjectMixinDictType) -> None:
        """Modify existing object based on a matching dictionary of properties.

        Args:
            obj_to_modify(DataObjectMixinDictType): dictionary containing attributes to match
            in the existing object.
            new_properties(DataObjectMixinDictType): properties to switch to in the new object.
        """
        raise NotImplementedError("Implement this in the derived class")

    @property
    @abstractmethod
    def table_header(self) -> str:
        """Returns table header as a string."""
        raise NotImplementedError("Implement this in the derived class")

    @property
    @abstractmethod
    def table_footer(self) -> str:
        """Returns table footer as a string."""
        raise NotImplementedError("Implement this in the derived class")

    @property
    @abstractmethod
    def _network_element_name(self) -> str:
        raise NotImplementedError("Implement this in the derived class")

    @staticmethod
    def resolve_carried_over_attributes(objects_to_resolve: Sequence[T]) -> Sequence[T]:
        """Resolves carried over attributes from previous time steps to the future timesteps.

        Args:
            objects_to_resolve (Sequence[DataObjectMixin]): list of objects to resolve carried over attributes for.
            Must be of homogenous type.
        """
        resolved_objects: list[T] = []

        sorted_by_date, unique_names = NetworkOperationsMixIn.__sort_by_date_name_sim_order(objects_to_resolve)

        # resolve by name
        for name in unique_names:
            matching_names = [x for x in sorted_by_date if x.name == name]
            resolving_by_name = NetworkOperationsMixIn.resolve_same_named_objects(matching_names)
            resolved_objects.extend(resolving_by_name)

        return resolved_objects

    @staticmethod
    def __sort_by_date_name_sim_order(objects_to_resolve: Sequence[T]) -> tuple[Sequence[T], list[str | None]]:
        """Sorts objects by date and name.

        Args:
            objects_to_resolve (Sequence[DataObjectMixin]): list of objects to resolve carried over attributes for.
            Must be of homogenous type.
        """
        if (len(objects_to_resolve) > 0 and
                not all(isinstance(x, type(objects_to_resolve[0])) for x in objects_to_resolve)):
            raise ValueError("Objects to resolve must be of the same type.")
        # order by date and the order entered in the simulator.
        current_ordering = list(enumerate(objects_to_resolve))
        sorted_by_date_sim_ordering = sorted(current_ordering, key=lambda x: (x[1].iso_date, x[0]))
        sorted_by_date = [x[1] for x in sorted_by_date_sim_ordering]
        # split by the name of the object
        unique_names = list({x.name: x for x in sorted_by_date}.keys())
        return sorted_by_date, unique_names

    @staticmethod
    def resolve_same_named_objects(sorted_by_date: Sequence[T]) -> Sequence[T]:
        """Resolves a subset of objects by date.

        Args:
            sorted_by_date (Sequence[DataObjectMixin]): list of objects to resolve carried over attributes for.
            Must be of homogenous type.

        Returns:
            Sequence[DataObjectMixin]: list of resolved objects with carried over attributes.

        """
        resolved_objects: list[T] = []
        for unresolved_obj in sorted_by_date:
            # append the first
            if len(resolved_objects) == 0:
                resolved_objects.append(unresolved_obj)
                continue
            last_resolved_object = resolved_objects[-1]

            new_resolved_object = copy.deepcopy(unresolved_obj)

            for attr, value in last_resolved_object.__dict__.items():
                if value is not None and getattr(unresolved_obj, attr, None) is None:
                    setattr(new_resolved_object, attr, value)
            resolved_objects.append(new_resolved_object)
        return resolved_objects

    @staticmethod
    def resolve_same_named_objects_constraints(sorted_by_data: Sequence[NexusConstraint]) -> Sequence[NexusConstraint]:
        """Resolves a subset of objects by date and applies clears in the constraints.

        Args:
            sorted_by_data (Sequence[NexusConstraint]): Sequence of constraints to resolve in .
        """
        resolved_objects: list[NexusConstraint] = []
        for unresolved_obj in sorted_by_data:
            # append the first
            if len(resolved_objects) == 0:
                resolved_objects.append(unresolved_obj)
                continue
            last_resolved_object = resolved_objects[-1]
            # copy to prevent overriding previously resolved objects
            last_resolved_copy = copy.deepcopy(last_resolved_object)
            new_resolved_object = copy.deepcopy(unresolved_obj)

            # collect which attributes to clear:
            clear_constraints_dict = {}
            if unresolved_obj.clear_q:
                clear_constraints_dict.update(unresolved_obj.get_rate_constraints_map())
            if unresolved_obj.clear_p:
                clear_constraints_dict.update(unresolved_obj.get_pressure_constraints_map())
            if unresolved_obj.clear_limit:
                clear_constraints_dict.update(unresolved_obj.get_limit_constraints_map())
            if unresolved_obj.clear_alq:
                clear_constraints_dict.update(unresolved_obj.get_alq_constraints_map())
            if unresolved_obj.clear_all:
                clear_constraints_dict.update(unresolved_obj.get_rate_constraints_map())
                clear_constraints_dict.update(unresolved_obj.get_pressure_constraints_map())
                clear_constraints_dict.update(unresolved_obj.get_limit_constraints_map())
                clear_constraints_dict.update(unresolved_obj.get_alq_constraints_map())
            # Clear them by setting to None on the copied previously resolved object
            for (clear_attr, _) in clear_constraints_dict.values():
                setattr(last_resolved_copy, clear_attr, None)

            skip_attributes = ['id', 'date', 'name', 'iso_date', 'clear_q', 'clear_p', 'clear_limit', 'clear_alq',
                               'clear_all', 'convert_qmult_to_reservoir_barrels', 'active_node']
            for attr, value in last_resolved_copy.__dict__.items():
                if value is None or attr in skip_attributes:
                    continue
                if getattr(unresolved_obj, attr, None) is None:
                    setattr(new_resolved_object, attr, value)

            # resolve convert_qmult_to_reservoir_barrels
            if last_resolved_object.convert_qmult_to_reservoir_barrels:
                overriding_qmult_constraints = [unresolved_obj.max_surface_gas_rate,
                                                unresolved_obj.max_surface_water_rate,
                                                unresolved_obj.max_surface_oil_rate,
                                                ]
                if any(x for x in overriding_qmult_constraints if x is not None):
                    # clear any qmult constraints if any of the surface rate constraints for oil, gas, water are set
                    new_resolved_object.convert_qmult_to_reservoir_barrels = None
                    new_resolved_object.qmult_oil_rate = None
                    new_resolved_object.qmult_gas_rate = None
                    new_resolved_object.qmult_water_rate = None
                elif 'QALLRMAX' in clear_constraints_dict:
                    # if the object is cleared, set to None
                    new_resolved_object.convert_qmult_to_reservoir_barrels = None
                else:
                    new_resolved_object.convert_qmult_to_reservoir_barrels = True

            resolved_objects.append(new_resolved_object)
        return resolved_objects

    @property
    def resolved_network_objects(self) -> Sequence[DataObjectMixin]:
        """Returns the resolved network objects."""
        if not hasattr(self, '_resolved_network_objects'):
            self._resolved_network_objects = self.resolve_carried_over_attributes(self.get_all())
        return self._resolved_network_objects

    def to_string_for_date(self, date: ISODateTime) -> str:
        """Returns a string representation of the network object for the date."""
        printable_str = ''
        # get the required objects for the date
        network_objects_for_date: list[DataObjectMixin] = [x for x in self.resolved_network_objects if
                                                           x.iso_date == date]
        named_network_objects_for_date = {x.name: None for x in network_objects_for_date}
        # only pick the last object by name for the date
        network_objects_for_date = [[x for x in network_objects_for_date if x.name == y][-1]
                                    for y in named_network_objects_for_date.keys()]

        if not network_objects_for_date:
            return printable_str
        printable_str += f'{self.table_header}\n'
        # collect the required table column header:
        headers = get_column_headers_required(network_objects_for_date)
        printable_str += ' '.join(headers) + '\n'
        for network_object in network_objects_for_date:
            printable_str += f'{network_object.to_table_line(headers=headers)}'

        printable_str += f'{self.table_footer}\n'

        return printable_str



"""Functions for collecting tables from a Nexus file to populate Nexus objects."""
from __future__ import annotations

import fnmatch
from datetime import timedelta, time
from typing import Any, Optional

import ResSimpy.FileOperations.file_operations as fo
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.FileOperations.File import File
from ResSimpy.FileOperations.file_operations import get_next_value, get_expected_token_value
from ResSimpy.Nexus.DataModels.Network.NexusActivationChange import NexusActivationChange
from ResSimpy.Nexus.DataModels.Network.NexusConLists import NexusConLists
from ResSimpy.Nexus.DataModels.Network.NexusNodeLists import NexusNodeLists
from ResSimpy.Nexus.DataModels.Network.NexusConstraint import NexusConstraint
from ResSimpy.Nexus.DataModels.Network.NexusWellLists import NexusWellLists
from ResSimpy.Nexus.DataModels.NexusWellList import NexusWellList
from ResSimpy.Nexus.NexusEnums.ActivationChangeEnum import ActivationChangeEnum
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Nexus.nexus_constraint_operations import load_inline_constraints
from ResSimpy.Nexus.nexus_file_operations import check_property_in_line, check_token, check_list_tokens, \
    load_table_to_objects
from ResSimpy.Nexus.nexus_load_list_table import load_table_to_lists
from ResSimpy.Time.ISODateTime import ISODateTime


# TODO refactor the collection of tables to an object with proper typing
def collect_all_tables_to_objects(nexus_file: File, table_object_map: dict[str, Any], start_date: Optional[str],
                                  default_units: Optional[UnitSystem], date_format: DateFormat) -> \
        tuple[dict[str, list[Any]], dict[str, list[NexusConstraint]]]:
    """Loads all tables from a given file.

    Args:
    ----
        nexus_file (File): NexusFile representation of the file.
        table_object_map (dict[str, Storage_Object]): dictionary containing the name of the table as keys and \
                    the object type to store the data from each row into. Require objects to have a get_keyword_mapping\
                    function
        model: (NexusSimulator): main simulator object
        start_date (Optional[str]): The model start date.
        default_units (Optional[UnitSystem]): The default unit system the Nexus model is set to.
        date_format (Optional[DateFormat]): The date format of the object.

    Raises:
    ------
    TypeError: if the unit system found in the property check is not a valid enum UnitSystem.

    Returns:
    -------
    tuple[dict[str, list[DataObject]], dict[str, list[NexusConstraint]]]: a tuple of two dictionaries of lists of \
    DataObjects. The first element is a dictionary of lists of objects keyed with the NexusTable name associated with \
    the object. The second element is a dictionary of lists of NexusConstraints keyed with the well name associated \
    with the constraint.
    """
    current_date = start_date
    nexus_object_results: dict[str, list[Any]] = {x: [] for x in table_object_map}
    nexus_constraints: dict[str, list[NexusConstraint]] = {}

    file_as_list: list[str] = nexus_file.get_flat_list_str_file
    table_start: int = -1
    table_end: int = -1
    property_dict: dict = {}
    token_found: Optional[str] = None
    network_names: list[str] = []
    well_names: list[str] = []
    well_lists: list[NexusWellList] = []
    is_deactivate_block = False
    is_activate_block = False
    is_actions_block = False
    is_constraints_block = False
    default_crossflow: Optional[str] = None
    default_shutin: Optional[str] = None
    for index, line in enumerate(file_as_list):
        # check for changes in unit system
        check_property_in_line(line, property_dict, file_as_list)
        unit_system = property_dict.get('UNIT_SYSTEM', default_units)
        if not isinstance(unit_system, UnitSystem):
            raise TypeError(f"Value found for {unit_system=} of type {type(unit_system)} \
                                not compatible, expected type UnitSystem Enum")

        if current_date is None:
            raise ValueError("Unable to determine date for object")

        # Important!: DEACTIVATE can also appear in the ACTIONS table in the surface network and CONSTRAINTS tables.
        # If we are already in one of those blocks, we don't want to consider it a new activate / deactivate block.
        if fo.check_token('ACTIONS', line=line):
            is_actions_block = True
        if fo.check_token('ENDACTIONS', line=line):
            is_actions_block = False
        if fo.check_token('CONSTRAINTS', line=line):
            is_constraints_block = True
        if fo.check_token('ENDCONSTRAINTS', line=line):
            is_constraints_block = False

        if not is_actions_block and not is_constraints_block:
            is_activate_block, is_deactivate_block, should_continue = \
                __activate_deactivate_checks(line=line,
                                             is_activate_block=is_activate_block,
                                             is_deactivate_block=is_deactivate_block,
                                             current_date=current_date,
                                             start_date=start_date,
                                             date_format=date_format,
                                             nexus_object_results=nexus_object_results)

            if should_continue:
                # If we have already read in all the values for this line in the previous call, continue onto the next
                # line.
                continue

        if check_token('TIME', line) and table_start < 0:
            time_value = get_expected_token_value(
                token='TIME', token_line=line, file_list=file_as_list,
                custom_message=f"Cannot find the date associated with the TIME card in {line=} at line number {index}")
            if time_value.upper() != 'PLUS':
                current_date = time_value
            else:
                plus_value = get_expected_token_value(token='PLUS', token_line=line, file_list=file_as_list,
                                                      custom_message="Cannot find the date associated with the TIME "
                                                                     f"PLUS card in {line=} at line number {index}")

                if current_date is None:
                    raise ValueError("Cannot calculate PLUS date without access to the initial date.")
                else:
                    new_datetime = ISODateTime.convert_to_iso(
                        current_date, date_format, start_date) + timedelta(days=float(plus_value))

                if date_format == DateFormat.DD_MM_YYYY:
                    if new_datetime.time() == time(0, 0, 0, 0):
                        current_date = new_datetime.strftime('%d/%m/%Y')
                    else:
                        current_date = new_datetime.strftime('%d/%m/%Y(%H:%M:%S)')
                elif date_format == DateFormat.MM_DD_YYYY:
                    if new_datetime.time() == time(0, 0, 0, 0):
                        current_date = new_datetime.strftime('%m/%d/%Y')
                    else:
                        current_date = new_datetime.strftime('%m/%d/%Y(%H:%M:%S)')
            continue

        if table_start < 0:
            token_found = check_list_tokens(list(table_object_map.keys()), line)
            if token_found is None or check_token('WELLCONTROL', line):
                default_crossflow, default_shutin = __set_crossflow_and_shutin_defaults(default_crossflow,
                                                                                        default_shutin,
                                                                                        file_as_list, index, line)
                continue
            # if a token is found get the starting index of the table
            table_start = index + 1
        if token_found is None:
            continue
        token_found = token_found.upper()

        if table_start > 0 and check_token('END' + token_found, line):
            table_end = index
        # if we have a complete table to read in start reading it into objects
        if 0 < table_start <= table_end:
            # cover for the case where we aren't currently reading in the table to an object.
            # if no object is provided by the map for the token found then skip the keyword and reset the tracking vars
            if table_object_map[token_found] is None:
                table_start = -1
                table_end = -1
                token_found = None
                continue
            list_objects = None
            property_map = table_object_map[token_found].get_keyword_mapping()
            if token_found == 'CONSTRAINTS':
                load_inline_constraints(file_as_list=file_as_list[table_start:table_end],
                                        constraint=table_object_map[token_found],
                                        current_date=current_date,
                                        unit_system=unit_system,
                                        property_map=property_map,
                                        existing_constraints=nexus_constraints,
                                        nexus_file=nexus_file,
                                        start_line_index=table_start,
                                        network_names=network_names,
                                        date_format=date_format,
                                        welllists=well_lists,
                                        start_date=start_date)

            elif token_found == 'QMULT' or token_found == 'CONSTRAINT':
                list_objects = load_table_to_objects(file_as_list=file_as_list[table_start:table_end],
                                                     row_object=table_object_map[token_found],
                                                     property_map=property_map,
                                                     current_date=current_date,
                                                     unit_system=unit_system,
                                                     constraint_obj_dict=nexus_constraints,
                                                     preserve_previous_object_attributes=True,
                                                     date_format=date_format,
                                                     welllists=well_lists,
                                                     start_date=start_date)

            elif token_found in [
                NexusWellLists.table_header(),
                NexusConLists.table_header(),
                NexusNodeLists.table_header(),
            ]:
                list_objects = load_table_to_lists(file_as_list=file_as_list[table_start - 1:table_end],
                                                   row_object=table_object_map[token_found],
                                                   table_header=token_found,
                                                   current_date=current_date,
                                                   previous_lists=nexus_object_results[token_found],
                                                   date_format=date_format)

                # If there is already a matching Welllist for this timestamp, update that with the additional changes,
                # rather than adding another one.
                existing_welllist = next(
                    (x for x in nexus_object_results[NexusWellLists.table_header()] if
                     x.name == list_objects[0][0].name and
                     x.date == list_objects[0][0].date), None)

                if existing_welllist is not None:
                    existing_welllist.elements_in_the_list = list_objects[0][0].wells
                    list_objects = []
                if token_found == NexusWellLists.table_header():
                    well_lists = [x[0] for x in list_objects]

            else:
                list_objects = load_table_to_objects(file_as_list=file_as_list[table_start:table_end],
                                                     row_object=table_object_map[token_found],
                                                     property_map=property_map,
                                                     current_date=current_date,
                                                     unit_system=unit_system, date_format=date_format,
                                                     well_names=well_names, start_date=start_date)

            # store objects found into right dictionary
            list_of_token_obj = nexus_object_results[token_found]
            # This statement ensures that CONSTRAINTs that are found in tables are actually added to the dictionary
            # under the same key as constraints to preserve their order
            if (token_found == 'CONSTRAINT' or token_found == 'QMULT') and list_objects is not None:
                for constraint, id_index in list_objects:
                    if isinstance(constraint, NexusConstraint):
                        obj_id = constraint.id
                        well_name = constraint.name
                    else:
                        obj_id = constraint
                        well_name = None
                    correct_line_index = id_index + table_start
                    nexus_file.add_object_locations(obj_id, [correct_line_index])
                    if well_name is None:
                        continue
                    if nexus_constraints.get(well_name, None) is not None:
                        nexus_constraints[well_name].append(constraint)
                    else:
                        nexus_constraints[well_name] = [constraint]

            elif list_objects is not None and isinstance(list_of_token_obj, list):
                list_of_token_obj.extend([x[0] for x in list_objects])
                # add the names from the nodes into the network names for wildcards
                # wildcards do not apply to LIST type nexus objects like WELLLIST.
                if 'LIST' not in token_found:
                    network_names.extend([x.name for x in list_of_token_obj])
                if token_found == 'WELLS':
                    well_names.extend([x.name for x in list_of_token_obj if x.name not in well_names])
                for new_object, id_index in list_objects:
                    correct_line_index = id_index + table_start
                    # temporary try statement until all objects have an id property
                    try:
                        obj_id = new_object.id
                        nexus_file.add_object_locations(obj_id, [correct_line_index])
                    except AttributeError:
                        pass
            # reset indices for further tables
            table_start = -1
            table_end = -1
            token_found = None

        if 'WELLS' in nexus_object_results.keys():
            __apply_default_connection_values(nexus_object_results=nexus_object_results,
                                              default_crossflow=default_crossflow, default_shutin=default_shutin)

    return nexus_object_results, nexus_constraints


def __set_crossflow_and_shutin_defaults(default_crossflow: Optional[str], default_shutin: Optional[str],
                                        file_as_list: list[str], index: int, line: str) -> tuple[Optional[str],
                                                                                                 Optional[str]]:
    """Sets the default values for Crossflow and Shutin."""
    if check_token(token='CROSSFLOW', line=line):
        crossflow_value = get_expected_token_value(token='CROSSFLOW', token_line=line,
                                                   file_list=file_as_list[index:])

        default_crossflow = crossflow_value
    if check_token(token='SHUTINON', line=line):
        default_shutin = 'ON'
    elif check_token(token='SHUTINOFF', line=line):
        default_shutin = 'OFF'
    elif check_token(token='SHUTIN_CELLGRAD', line=line):
        default_shutin = 'CELLGRAD'
    return default_crossflow, default_shutin


def __apply_default_connection_values(nexus_object_results: dict[str, list[Any]], default_crossflow: Optional[str],
                                      default_shutin: Optional[str]) -> None:
    """Applies the values for Crossflow from the defaults if they aren't set separately in a table."""
    all_connections = nexus_object_results['WELLS'] + nexus_object_results['GASWELLS']
    for connection in all_connections:
        if connection.crossflow is None:
            connection.crossflow = default_crossflow
        if connection.crossshut is None:
            connection.crossshut = default_shutin


def __activate_deactivate_checks(line: str, is_activate_block: bool,
                                 is_deactivate_block: bool, current_date: str, start_date: str | None,
                                 date_format: DateFormat, nexus_object_results: dict[str, list[Any]]) \
        -> tuple[bool, bool, bool]:
    """Checks if the line being read is the start of an activate / deactivate block, and if it is, sets the property on
    the related Wellconnection objects found.

    Returns:
    tuple[bool, bool, bool]: a tuple of three bools, stating whether the following text is an activate block, whether \
        it is a deactivate block, and whether the loop calling this method should continue to the next line.
    """
    # Handle activation / deactivation of well connections
    if check_token('DEACTIVATE', line):
        is_deactivate_block = True
        return is_activate_block, is_deactivate_block, True

    if check_token('ACTIVATE', line):
        is_activate_block = True
        return is_activate_block, is_deactivate_block, True

    if not is_activate_block and not is_deactivate_block:
        return is_activate_block, is_deactivate_block, False

    if check_token('CONNECTION', line):
        return is_activate_block, is_deactivate_block, True

    if check_token('ENDDEACTIVATE', line) or check_token('ENDACTIVATE', line):
        is_deactivate_block = False
        is_activate_block = False
        return is_activate_block, is_deactivate_block, True

    # loop through all the connection names found in the block, and create the ActivationChange objects.
    found_connections = []
    affected_connection_name = get_next_value(start_line_index=0, file_as_list=[line])

    # Get a list of all the distinct connection names that might be affected.
    all_well_connection_names = []
    connection_types = {'WELLS', 'GASWELLS', 'NODECON'} & set(nexus_object_results)
    for connection_type in connection_types:
        all_well_connection_names.extend([connection.name for
                                          connection in nexus_object_results[connection_type]])
    all_well_connections = list(dict.fromkeys(all_well_connection_names))

    while affected_connection_name is not None:
        change = ActivationChangeEnum.ACTIVATE if is_activate_block else ActivationChangeEnum.DEACTIVATE
        # If a wildcard is found, ensure that all connections that match the wildcard are dealt with.
        if '*' in affected_connection_name:
            all_affected_well_connections = [x for x in all_well_connections if
                                             fnmatch.fnmatch(x.casefold(), affected_connection_name.casefold())]

            # Create an object for each affected connection
            for connection in all_affected_well_connections:
                new_activation_change = NexusActivationChange(change=change, name=connection,
                                                              date=current_date,
                                                              start_date=start_date, date_format=date_format)
                nexus_object_results['ACTIVATE_DEACTIVATE'].append(new_activation_change)

        else:
            new_activation_change = NexusActivationChange(change=change, name=affected_connection_name,
                                                          date=current_date, start_date=start_date,
                                                          date_format=date_format)
            nexus_object_results['ACTIVATE_DEACTIVATE'].append(new_activation_change)

        found_connections.append(affected_connection_name)
        # Move to the next value
        affected_connection_name = get_next_value(start_line_index=0, file_as_list=[line],
                                                  ignore_values=found_connections)

    return is_activate_block, is_deactivate_block, False



"""This module contains the NexusConstraint class, which is used to represent a constraint in the Nexus model."""
from __future__ import annotations

import warnings
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.Constraint import Constraint
from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat


@dataclass(repr=False)
class NexusConstraint(Constraint):
    """Class representing a single constraint object within the NexusNetwork for a single datetime.

    Attributes:
    name (str): name of the well (NAME).
    max_surface_oil_rate (float): max surface oil rate (QOSMAX).
    max_surface_gas_rate (float): max surface gas rate (QGSMAX).
    max_surface_water_rate (float): max surface (QWSMAX).
    max_surface_liquid_rate (float): max surface liquid rate (QLIQSMAX).
    max_hc_molar_rate (float): Max hc molar rate (QMHCMAX).
    max_reverse_surface_oil_rate (float): max reverse surface oil rate (QOSMAX-).
    max_reverse_surface_gas_rate (float): max reverse surface gas rate (QGSMAX-).
    max_reverse_surface_water_rate (float): max reverse surface water rate (QWSMAX-).
    max_reverse_surface_liquid_rate (float): max reverse surface liquid rate (QLIQSMAX-).
    max_reservoir_oil_rate (float): max reservoir oil rate (QOMAX).
    max_reservoir_gas_rate (float): max reservoir gas rate (QGMAX).
    max_reservoir_water_rate (float): max reservoir water rate (QWMAX).
    max_reservoir_liquid_rate (float): max reservoir liquid rate (QLIQMAX).
    max_reservoir_total_fluids_rate (float): max reservoir total fluids rate (QALLMAX).
    max_reservoir_hc_rate (float): max reservoir hc rate (QHCMAX).
    max_reverse_reservoir_oil_rate (float): max reverse reservoir oil rate (QOMAX-).
    max_reverse_reservoir_gas_rate (float): max reverse reservoir gas rate (QGMAX-).
    max_reverse_reservoir_water_rate (float): max reverse reservoir water rate (QWMAX-).
    max_reverse_reservoir_liquid_rate (float): max reverse reservoir liquid rate (QLIQMAX-).
    max_reverse_reservoir_total_fluids_rate (float): max reverse reservoir total fluids rate (QALLMAX-).
    max_reverse_reservoir_hc_rate (float): max reverse reservoir hc rate (QHCMAX-).
    min_pressure (float): min pressure (PMIN).
    max_pressure (float): max pressure (PMAX).
    max_wag_water_pressure (float): max wag water pressure (PWMAX).
    max_wag_gas_pressure (float): max wag gas pressure (PGMAX).
    bottom_hole_pressure (float): bottom hole pressure (BHP).
    tubing_head_pressure (float): tubing head pressure (THP).
    min_surface_oil_rate (float): min surface oil rate (QOSMIN).
    min_surface_gas_rate (float): min surface gas rate (QGSMIN).
    min_surface_water_rate (float): min surface water rate (QWSMIN).
    min_surface_liquid_rate (float): min surface liquid rate (QLIQSMIN).
    min_reservoir_oil_rate (float): min reservoir oil rate (QOMIN).
    min_reservoir_gas_rate (float): min reservoir gas rate (QGMIN).
    min_reservoir_water_rate (float): min reservoir water rate (QWMIN).
    min_reservoir_liquid_rate (float): min reservoir liquid rate (QLIQMIN).
    min_reservoir_total_fluids_rate (float): min reservoir total fluids rate (QALLMIN).
    min_reservoir_hc_rate (float): min reservoir hc rate (QHCMIN).
    min_reverse_surface_oil_rate (float): min reservoir oil rate (QOSMIN-).
    min_reverse_surface_gas_rate (float): min reservoir gas rate (QGSMIN-).
    min_reverse_surface_water_rate (float): min reservoir water rate (QWSMIN-).
    min_reverse_surface_liquid_rate (float): min reservoir liquid rate (QLIQSMIN-).
    min_reverse_reservoir_oil_rate (float): min reverse reservoir oil rate (QOMIN-).
    min_reverse_reservoir_gas_rate (float): min reverse reservoir gas rate (QGMIN-).
    min_reverse_reservoir_water_rate (float): min reverse reservoir water rate (QWMIN-).
    min_reverse_reservoir_liquid_rate (float): min reverse reservoir liquid rate (QLIQMIN-).
    min_reverse_reservoir_total_fluids_rate (float): min reverse reservoir total fluids rate (QALLMIN-).
    min_reverse_reservoir_hc_rate (float): min reverse reservoir hc rate (QHCMIN-).
    max_watercut (float): max watercut (WCUTMAX).
    max_watercut_plug (float): max watercut plug (WCUTPLUG).
    max_watercut_plugplus (float): max watercut plugplus (WCUTPLUGPLUS).
    max_watercut_perf (float): max watercut perf (WCUTPERF).
    max_watercut_perfplus (float): max watercut perfplus (WCUTPERFPLUS).
    max_wor (float): max wor (WORMAX).
    max_wor_plug (float): max wor plug (WORPLUG).
    max_wor_plug_plus (float): max wor plug plus (WORPLUGPLUS).
    max_wor_perf (float): max wor perf (WORPERF).
    max_wor_perfplus (float): max wor perfplus (WORPERFPLUS).
    max_gor (float): max gor (GORMAX).
    max_gor_plug (float): max gor plug (GORPLUG).
    max_gor_plug_plus (float): max gor plug plus (GORPLUGPLUS).
    max_gor_perf (float): max gor perf (GORPERF).
    max_gor_perfplus (float): max gor perfplus (GORPERFPLUS).
    max_lgr (float): max lgr (LGRMAX).
    max_lgr_plug (float): max lgr plug (LGRPLUG).
    max_lgr_plug_plus (float): max lgr plug plus (LGRPLUGPLUS).
    max_lgr_perf (float): max lgr perf (LGRPERF).
    max_lgr_perfplus (float): max lgr perfplus (LGRPERFPLUS).
    max_cum_gas_prod (float): max cum gas prod (CGLIM).
    max_cum_water_prod (float): max cum water prod (CWLIM).
    max_cum_oil_prod (float): max cum oil prod (COLIM).

    artificial_lift_number (int): artificial lift number within the hydraulic table (ALQ)
    max_choke_setting (float): maximum choke/ICD/valve setting (SETTING)
    min_gas_lift_efficiency (float): minimum gas lift efficiency below which the connection will be shut in(GLEFMIN)
    gl_additive_correction (float): additive correction term to value of gas/liquid ratio from optimal gl \
    tables (GLRADD)
    active_node (bool): active/inactive node/well (ACTIVATE)
    pump_power (float): power for the pump (POWER)
    pump_speed (float): maximum speed of the pump/esp/compressor (SPEED)
    choke_limit (str): ON/OFF for whether the esp should exceed the choke limit (CHOKELIMIT)
    manifold_position (int): position in the manifold for the well (POSITION)
    """

    well_name: Optional[str] = None
    max_hc_molar_rate: Optional[float] = None
    max_reverse_surface_oil_rate: Optional[float] = None
    max_reverse_surface_gas_rate: Optional[float] = None
    max_reverse_surface_water_rate: Optional[float] = None
    max_reverse_surface_liquid_rate: Optional[float] = None
    max_reservoir_hc_rate: Optional[float] = None
    max_reverse_reservoir_oil_rate: Optional[float] = None
    max_reverse_reservoir_gas_rate: Optional[float] = None
    max_reverse_reservoir_water_rate: Optional[float] = None
    max_reverse_reservoir_liquid_rate: Optional[float] = None
    max_reverse_reservoir_total_fluids_rate: Optional[float] = None
    max_reverse_reservoir_hc_rate: Optional[float] = None
    max_avg_comp_dp: Optional[float] = None
    max_comp_dp: Optional[float] = None
    max_comp_dp_by_zone: Optional[dict[int, float]] = None
    min_pressure: Optional[float] = None
    max_pressure: Optional[float] = None
    max_wag_water_pressure: Optional[float] = None
    max_wag_gas_pressure: Optional[float] = None
    min_reverse_surface_oil_rate: Optional[float] = None
    min_reverse_surface_gas_rate: Optional[float] = None
    min_reverse_surface_water_rate: Optional[float] = None
    min_reverse_surface_liquid_rate: Optional[float] = None
    min_reservoir_oil_rate: Optional[float] = None
    min_reservoir_gas_rate: Optional[float] = None
    min_reservoir_water_rate: Optional[float] = None
    min_reservoir_liquid_rate: Optional[float] = None
    min_reservoir_total_fluids_rate: Optional[float] = None
    min_reservoir_hc_rate: Optional[float] = None
    min_reverse_reservoir_oil_rate: Optional[float] = None
    min_reverse_reservoir_gas_rate: Optional[float] = None
    min_reverse_reservoir_water_rate: Optional[float] = None
    min_reverse_reservoir_liquid_rate: Optional[float] = None
    min_reverse_reservoir_total_fluids_rate: Optional[float] = None
    min_reverse_reservoir_hc_rate: Optional[float] = None
    gor_limit_exponent: Optional[float] = None
    gor_limit_frequency: Optional[float] = None
    max_cum_gas_prod: Optional[float] = None
    max_cum_water_prod: Optional[float] = None
    max_cum_oil_prod: Optional[float] = None
    max_qmult_total_reservoir_rate: Optional[float] = None
    convert_qmult_to_reservoir_barrels: Optional[bool] = None
    qmult_oil_rate: Optional[float] = None
    qmult_water_rate: Optional[float] = None
    qmult_gas_rate: Optional[float] = None
    use_qmult_qoil_surface_rate: Optional[bool] = None
    use_qmult_qwater_surface_rate: Optional[bool] = None
    use_qmult_qgas_surface_rate: Optional[bool] = None
    use_qmult_qoilqwat_surface_rate: Optional[bool] = None
    artificial_lift_number: Optional[int] = None
    max_choke_setting: Optional[float] = None
    min_gas_lift_efficiency: Optional[float] = None
    gl_additive_correction: Optional[float] = None
    active_node: Optional[bool] = None
    pump_power: Optional[float] = None
    pump_speed: Optional[float] = None
    choke_limit: Optional[str] = None
    manifold_position: Optional[int] = None
    clear_all: Optional[bool] = None
    clear_q: Optional[bool] = None
    clear_limit: Optional[bool] = None
    clear_alq: Optional[bool] = None
    clear_p: Optional[bool] = None

    def __init__(self, properties_dict: Optional[DataObjectMixinDictType] = None,
                 date: Optional[str] = None, date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None, name: Optional[str] = None, well_name: Optional[str] = None,
                 max_surface_oil_rate: Optional[float] = None, max_surface_gas_rate: Optional[float] = None,
                 max_surface_water_rate: Optional[float] = None, max_surface_liquid_rate: Optional[float] = None,
                 max_hc_molar_rate: Optional[float] = None, max_reverse_surface_oil_rate: Optional[float] = None,
                 max_reverse_surface_gas_rate: Optional[float] = None,
                 max_reverse_surface_water_rate: Optional[float] = None,
                 max_reverse_surface_liquid_rate: Optional[float] = None,
                 max_reservoir_oil_rate: Optional[float] = None, max_reservoir_gas_rate: Optional[float] = None,
                 max_reservoir_water_rate: Optional[float] = None, max_reservoir_liquid_rate: Optional[float] = None,
                 max_reservoir_hc_rate: Optional[float] = None, max_reverse_reservoir_oil_rate: Optional[float] = None,
                 max_reverse_reservoir_gas_rate: Optional[float] = None,
                 max_reverse_reservoir_water_rate: Optional[float] = None,
                 max_reverse_reservoir_liquid_rate: Optional[float] = None,
                 max_reverse_reservoir_total_fluids_rate: Optional[float] = None,
                 max_reverse_reservoir_hc_rate: Optional[float] = None, max_avg_comp_dp: Optional[float] = None,
                 max_comp_dp: Optional[float] = None, min_pressure: Optional[float] = None,
                 max_pressure: Optional[float] = None, max_wag_water_pressure: Optional[float] = None,
                 max_wag_gas_pressure: Optional[float] = None, min_reverse_surface_oil_rate: Optional[float] = None,
                 min_reverse_surface_gas_rate: Optional[float] = None,
                 min_reverse_surface_water_rate: Optional[float] = None,
                 min_reverse_surface_liquid_rate: Optional[float] = None,
                 min_reservoir_oil_rate: Optional[float] = None, min_reservoir_gas_rate: Optional[float] = None,
                 min_reservoir_water_rate: Optional[float] = None, min_reservoir_liquid_rate: Optional[float] = None,
                 min_reservoir_total_fluids_rate: Optional[float] = None, min_reservoir_hc_rate: Optional[float] = None,
                 min_reverse_reservoir_oil_rate: Optional[float] = None,
                 min_reverse_reservoir_gas_rate: Optional[float] = None,
                 min_reverse_reservoir_water_rate: Optional[float] = None,
                 min_reverse_reservoir_liquid_rate: Optional[float] = None,
                 min_reverse_reservoir_total_fluids_rate: Optional[float] = None,
                 min_reverse_reservoir_hc_rate: Optional[float] = None, gor_limit: Optional[float] = None,
                 gor_limit_exponent: Optional[float] = None, gor_limit_frequency: Optional[float] = None,
                 max_cum_gas_prod: Optional[float] = None, max_cum_water_prod: Optional[float] = None,
                 max_cum_oil_prod: Optional[float] = None, max_qmult_total_reservoir_rate: Optional[float] = None,
                 convert_qmult_to_reservoir_barrels: Optional[bool] = None, qmult_oil_rate: Optional[float] = None,
                 qmult_water_rate: Optional[float] = None, qmult_gas_rate: Optional[float] = None,
                 use_qmult_qoil_surface_rate: Optional[bool] = None,
                 use_qmult_qwater_surface_rate: Optional[bool] = None,
                 use_qmult_qgas_surface_rate: Optional[bool] = None,
                 use_qmult_qoilqwat_surface_rate: Optional[bool] = None, artificial_lift_number: Optional[int] = None,
                 max_choke_setting: Optional[float] = None, min_gas_lift_efficiency: Optional[float] = None,
                 gl_additive_correction: Optional[float] = None, active_node: Optional[bool] = None,
                 pump_power: Optional[float] = None, pump_speed: Optional[float] = None,
                 choke_limit: Optional[str] = None, mainfold_position: Optional[int] = None,
                 clear_all: Optional[bool] = None, clear_q: Optional[bool] = None, clear_limit: Optional[bool] = None,
                 clear_alq: Optional[bool] = None, clear_p: Optional[bool] = None,
                 min_surface_oil_rate: Optional[float] = None, min_surface_gas_rate: Optional[float] = None,
                 min_surface_water_rate: Optional[float] = None, min_surface_liquid_rate: Optional[float] = None,
                 bottom_hole_pressure: Optional[float] = None, tubing_head_pressure: Optional[float] = None,
                 max_reservoir_total_fluids_rate: Optional[float] = None, max_watercut: Optional[float] = None,
                 max_watercut_plug: Optional[float] = None, max_watercut_plugplus: Optional[float] = None,
                 max_watercut_perf: Optional[float] = None, max_watercut_perfplus: Optional[float] = None,
                 max_wor: Optional[float] = None, max_wor_plug: Optional[float] = None,
                 max_wor_plug_plus: Optional[float] = None, max_wor_perf: Optional[float] = None,
                 max_wor_perfplus: Optional[float] = None, max_gor: Optional[float] = None,
                 max_gor_plug: Optional[float] = None, max_gor_plug_plus: Optional[float] = None,
                 max_gor_perf: Optional[float] = None, max_gor_perfplus: Optional[float] = None,
                 max_lgr: Optional[float] = None, max_lgr_plug: Optional[float] = None,
                 max_lgr_plug_plus: Optional[float] = None, max_lgr_perf: Optional[float] = None,
                 max_lgr_perfplus: Optional[float] = None, max_comp_dp_by_zone: Optional[dict[int, float]] = None) \
            -> None:
        r"""Initialises the NexusConstraint class.

        Args:
            properties_dict (dict): dict of the properties to set on the object.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            name (Optional[str]): The name of the object.
            well_name (Optional[str]): The well name.
            max_surface_oil_rate (Optional[float]): Max surface oil rate (QOSMAX).
            max_surface_gas_rate (Optional[float]): Mamx surface gas rate (QGSMAX).
            max_surface_water_rate (Optional[float]): ax surface (QWSMAX).
            max_surface_liquid_rate (Optional[float]): Max surface liquid rate (QLIQSMAX).
            max_hc_molar_rate (Optional[float]): Max hc molar rate (QMHCMAX).
            max_reverse_surface_oil_rate (Optional[float]): Max reverse surface oil rate (QOSMAX-).
            max_reverse_surface_gas_rate (Optional[float]): Max reverse surface gas rate (QGSMAX-).
            max_reverse_surface_water_rate (Optional[float]): Max reverse surface water rate (QWSMAX-).
            max_reverse_surface_liquid_rate (Optional[float]): Max reverse surface liquid rate (QLIQSMAX-).
            max_reservoir_oil_rate (Optional[float]): Max reservoir oil rate (QOMAX).
            max_reservoir_gas_rate (Optional[float]): Max reservoir gas rate (QGMAX).
            max_reservoir_water_rate (Optional[float]): Max reservoir water rate (QWMAX).
            max_reservoir_liquid_rate (Optional[float]): Max reservoir liquid rate (QLIQMAX).
            max_reservoir_hc_rate (Optional[float]): Max reservoir hc rate (QHCMAX).
            max_reverse_reservoir_oil_rate (Optional[float]): Max reverse reservoir oil rate (QOMAX-).
            max_reverse_reservoir_gas_rate (Optional[float]): Max reverse reservoir gas rate (QGMAX-).
            max_reverse_reservoir_water_rate (Optional[float]): Max reverse reservoir water rate (QWMAX-).
            max_reverse_reservoir_liquid_rate (Optional[float]): Max reverse reservoir liquid rate (QLIMAX-).
            max_reverse_reservoir_total_fluids_rate (Optional[float]): max reverse reservoir total
            fluids rate (QALLMAX-).
            max_reverse_reservoir_hc_rate (Optional[float]): Max reverse reservoir hc rate (QHCMAX-).
            max_avg_comp_dp (Optional[float]): Specifies average drawdown/build up for well (DPBHAVG).
            max_comp_dp (Optional[float]): Specifies maximum drawdown/build up for well (DPBHMX).
            max_comp_dp_by_zone (Optional[dict[int, float]]): Specifies maximum drawdown/build up for well by zone
            (DPBHMX).
            min_pressure (Optional[float]): Min pressure (PMIN).
            max_pressure (Optional[float]): Max pressure (PMAX).
            max_wag_water_pressure (Optional[float]): Max wag water pressure (PWMAX).
            max_wag_gas_pressure (Optional[float]): Max wag gas pressure (PGMAX).
            min_reverse_surface_oil_rate (Optional[float]): Min reservoir oil rate (QOSMIN-).
            min_reverse_surface_gas_rate (Optional[float]): Min reservoir gas rate (QGSMIN-).
            min_reverse_surface_water_rate (Optional[float]): Min reservoir water rate (QWSMIN-).
            min_reverse_surface_liquid_rate (Optional[float]): Min reservoir liquid rate (QLIQSMIN-).
            min_reservoir_oil_rate (Optional[float]): Min reservoir oil rate (QOMIN).
            min_reservoir_gas_rate (Optional[float]): Min reservoir gas rate (QGMIN).
            min_reservoir_water_rate (Optional[float]): Min reservoir water rate (QWMIN).
            min_reservoir_liquid_rate (Optional[float]): Min reservoir liquid rate (QLIQMIN).
            min_reservoir_total_fluids_rate (Optional[float]): Min reservoir total fluids rate (QALLMIN).
            min_reverse_reservoir_total_fluids_rate (Optional[float]): Min reverse reservoir total
            fluids rate (QALLMIN-).
            min_reverse_reservoir_gas_rate (Optional[float]): Min reverse reservoir gas rate (QGMIN-).
            min_reverse_reservoir_liquid_rate(Optional[float]): Min reverse reservoir liquid rate (QLIQMIN-).
            min_reverse_reservoir_oil_rate (Optional[float]): Min reverse reservoir oil rate (QOMIN-).
            min_reverse_reservoir_water_rate (Optional[float]): Min reverse reservoir water rate (QWMIN-).
            min_reservoir_hc_rate (Optional[float]): Min reservoir hc rate (QHCMIN).
            min_reverse_reservoir_hc_rate (Optional[float]): Min reverse reservoir hc rate (QHCMIN-).
            gor_limit (Optional[float]): Specifies desired maximum gas/oil ration for a producing well (GORLIM).
            gor_limit_exponent (Optional[float]):  Specifies the exponent “exp” to be used in the GORLIM.
            gor_limit_exponent (Optional[float]):  Specifies the exponent “exp” to be used in the GORLIM.
            gor_limit_frequency (Optional[float]): Specifies the frequency of computation of new rate constraints
            when a well is affected by an active GORLIM constraint.
            max_cum_gas_prod (Optional[float]): Max cum gas prod (CGLIM).
            max_cum_water_prod (Optional[float]): Max cum water prod (CWLIM).
            max_cum_oil_prod (Optional[float]): Max cum oil prod (COLIM).
            max_qmult_total_reservoir_rate (Optional[float]): Limits flow rates in linear connecitons to a well with
            QMULT data (QALLRMAX).
            convert_qmult_to_reservoir_barrels (Optional[bool]): Surface rates input in QMULT table will be
            converted to reservoir barrel equivalents.
            qmult_oil_rate (Optional[float]): Specifies threshold surface oil production rate (QOIL).
            qmult_water_rate (Optional[float]): Specifies that water rate is used to determine the pressure drop
            across the valve (QWATER).
            qmult_gas_rate (Optional[float]): Specifies gas rate is used to determine the pressure drop across
            the valve (QGAS).
            use_qmult_qoil_surface_rate (Optional[bool]): The Qoil value will be used as a surface oil rate constraint
            (QOSMAX).
            use_qmult_qwater_surface_rate (Optional[bool]): The Qwater value will be used as a surface water
            rate constraint. (QWSMAX).
            use_qmult_qgas_surface_rate (Optional[bool]): The Qgas value will be used as surface gas rate constraint
            (QGSMAX).
            use_qmult_qoilqwat_surface_rate (Optional[bool]): The sum of Qoil and Qwater will be used as a surface
            liquid rate constraint (QLIQSMAX).
            artificial_lift_number (Optional[int]): Artificial lift number within the hydraulic table (ALQ).
            max_choke_setting (Optional[float]): Maximum choke/ICD/valve setting (SETTING).
            min_gas_lift_efficiency (Optional[float]): Minimum gas lift efficiency below which the connection
            will be shut in(GLEFMIN).
            gl_additive_correction (Optional[float]): Additive correction term to value of gas or liquid ratio from
            optimal gl\tables (GLRADD).
            active_node (Optional[bool]): active or inactive node or well (ACTIVATE).
            pump_power (Optional[float]): Power for the pump (POWER).
            pump_speed (Optional[float]): Maximum speed of the pump,esp,compressor (SPEED).
            choke_limit (Optional[str]): ON or OFF for whether the esp should exceed the choke limit (CHOKELIMIT).
            mainfold_position (Optional[int]): Position in the manifold for the well (POSITION).
            clear_all (Optional[bool]): Specifies previous procedures that have been input will be ignored (CLEAR).
            clear_q (Optional[bool]): Remove all rate constraints for the well
            clear_limit (Optional[bool]): Specifies that all limit constraints for this well
            will be removed (CLEARLIMIT).
            clear_alq (Optional[bool]): Specifies that the artificial lift quantity will be set
            to its default (CLEARALQ).
            clear_p (Optional[bool]): Remove all pressure constraints for the well (CLEARP).
            bottom_hole_pressure (float): bottom hole pressure (BHP).
            max_gor (float): max gor (GORMAX).
            max_gor_perf (float): max gor perf (GORPERF).
            max_gor_plug (float): max gor plug (GORPLUG).
            max_gor_plug_plus (float): max gor plug plus (GORPLUGPLUS).
            max_gor_perfplus (float): max gor perfplus (GORPERFPLUS).
            max_lgr (float): max lgr (LGRMAX).
            max_lgr_plug (float): max lgr plug (LGRPLUG).
            max_lgr_plug_plus (float): max lgr plug plus (LGRPLUGPLUS).
            max_lgr_perf (float): max lgr perf (LGRPERF).
            max_lgr_perfplus (float): max lgr perfplus (LGRPERFPLUS).
            max_watercut (float): max watercut (WCUTMAX).
            max_watercut_plug (float): max watercut plug (WCUTPLUG).
            max_watercut_plugplus (float): max watercut plugplus (WCUTPLUGPLUS).
            max_watercut_perf (float): max watercut perf (WCUTPERF).
            max_watercut_perfplus (float): max watercut perfplus (WCUTPERFPLUS).
            max_wor (float): max wor (WORMAX).
            max_wor_plug (float): max wor plug (WORPLUG).
            max_wor_plug_plus (float): max wor plug plus (WORPLUGPLUS).
            max_wor_perf (float): max wor perf (WORPERF).
            max_wor_perfplus (float): max wor perfplus (WORPERFPLUS).
            min_surface_oil_rate (float): min surface oil rate (QOSMIN).
            min_surface_gas_rate (float): min surface gas rate (QGSMIN).
            min_surface_water_rate (float): min surface water rate (QWSMIN).
            min_surface_liquid_rate (float): min surface liquid rate (QLIQSMIN).
            tubing_head_pressure (float): tubing head pressure (THP).
            max_reservoir_total_fluids_rate (float): max reservoir total fluids rate (QALLMAX).
        """

        self.well_name = well_name
        self.max_hc_molar_rate = max_hc_molar_rate
        self.max_reverse_surface_oil_rate = max_reverse_surface_oil_rate
        self.max_reverse_surface_gas_rate = max_reverse_surface_gas_rate
        self.max_reverse_surface_water_rate = max_reverse_surface_water_rate
        self.max_reverse_surface_liquid_rate = max_reverse_surface_liquid_rate
        self.max_reservoir_hc_rate = max_reservoir_hc_rate
        self.max_reverse_reservoir_oil_rate = max_reverse_reservoir_oil_rate
        self.max_reverse_reservoir_gas_rate = max_reverse_reservoir_gas_rate
        self.max_reverse_reservoir_water_rate = max_reverse_reservoir_water_rate
        self.max_reverse_reservoir_liquid_rate = max_reverse_reservoir_liquid_rate
        self.max_reverse_reservoir_total_fluids_rate = max_reverse_reservoir_total_fluids_rate
        self.max_reverse_reservoir_hc_rate = max_reverse_reservoir_hc_rate
        self.min_pressure = min_pressure
        self.max_pressure = max_pressure
        self.max_wag_water_pressure = max_wag_water_pressure
        self.max_wag_gas_pressure = max_wag_gas_pressure
        self.min_reverse_surface_oil_rate = min_reverse_surface_oil_rate
        self.min_reverse_surface_gas_rate = min_reverse_surface_gas_rate
        self.min_reverse_surface_water_rate = min_reverse_surface_water_rate
        self.min_reverse_surface_liquid_rate = min_reverse_surface_liquid_rate
        self.min_reservoir_oil_rate = min_reservoir_oil_rate
        self.min_reservoir_gas_rate = min_reservoir_gas_rate
        self.min_reservoir_water_rate = min_reservoir_water_rate
        self.min_reservoir_liquid_rate = min_reservoir_liquid_rate
        self.min_reservoir_total_fluids_rate = min_reservoir_total_fluids_rate
        self.min_reservoir_hc_rate = min_reservoir_hc_rate
        self.min_reverse_reservoir_oil_rate = min_reverse_reservoir_oil_rate
        self.min_reverse_reservoir_gas_rate = min_reverse_reservoir_gas_rate
        self.min_reverse_reservoir_water_rate = min_reverse_reservoir_water_rate
        self.min_reverse_reservoir_liquid_rate = min_reverse_reservoir_liquid_rate
        self.min_reverse_reservoir_total_fluids_rate = min_reverse_reservoir_total_fluids_rate
        self.min_reverse_reservoir_hc_rate = min_reverse_reservoir_hc_rate
        self.gor_limit_exponent = gor_limit_exponent
        self.gor_limit_frequency = gor_limit_frequency
        self.max_cum_gas_prod = max_cum_gas_prod
        self.max_cum_water_prod = max_cum_water_prod
        self.max_cum_oil_prod = max_cum_oil_prod
        self.max_qmult_total_reservoir_rate = max_qmult_total_reservoir_rate
        self.convert_qmult_to_reservoir_barrels = convert_qmult_to_reservoir_barrels
        self.qmult_oil_rate = qmult_oil_rate
        self.qmult_water_rate = qmult_water_rate
        self.qmult_gas_rate = qmult_gas_rate
        self.use_qmult_qoil_surface_rate = use_qmult_qoil_surface_rate
        self.use_qmult_qwater_surface_rate = use_qmult_qwater_surface_rate
        self.use_qmult_qgas_surface_rate = use_qmult_qgas_surface_rate
        self.use_qmult_qoilqwat_surface_rate = use_qmult_qoilqwat_surface_rate
        self.artificial_lift_number = artificial_lift_number
        self.max_choke_setting = max_choke_setting
        self.min_gas_lift_efficiency = min_gas_lift_efficiency
        self.gl_additive_correction = gl_additive_correction
        self.pump_power = pump_power
        self.pump_speed = pump_speed
        self.choke_limit = choke_limit
        self.mainfold_position = mainfold_position
        self.clear_all = clear_all
        self.clear_q = clear_q
        self.clear_limit = clear_limit
        self.clear_alq = clear_alq
        self.clear_p = clear_p
        self.max_comp_dp_by_zone = max_comp_dp_by_zone

        super().__init__(_date_format=date_format, _start_date=start_date, _unit_system=unit_system, name=name,
                         max_surface_oil_rate=max_surface_oil_rate,
                         max_surface_gas_rate=max_surface_gas_rate,
                         max_surface_water_rate=max_surface_water_rate, max_surface_liquid_rate=max_surface_liquid_rate,
                         max_reservoir_oil_rate=max_reservoir_oil_rate, max_reservoir_gas_rate=max_reservoir_gas_rate,
                         max_reservoir_water_rate=max_reservoir_water_rate,
                         max_reservoir_liquid_rate=max_reservoir_liquid_rate, max_avg_comp_dp=max_avg_comp_dp,
                         max_comp_dp=max_comp_dp, min_surface_oil_rate=min_surface_oil_rate,
                         min_surface_gas_rate=min_surface_gas_rate, min_surface_water_rate=min_surface_water_rate,
                         min_surface_liquid_rate=min_surface_liquid_rate, bottom_hole_pressure=bottom_hole_pressure,
                         tubing_head_pressure=tubing_head_pressure,
                         max_reservoir_total_fluids_rate=max_reservoir_total_fluids_rate, max_watercut=max_watercut,
                         max_watercut_plug=max_watercut_plug, max_watercut_plugplus=max_watercut_plugplus,
                         max_watercut_perf=max_watercut_perf, max_watercut_perfplus=max_watercut_perfplus,
                         max_wor=max_wor, max_wor_plug=max_wor_plug, max_wor_plug_plus=max_wor_plug_plus,
                         max_wor_perf=max_wor_perf, max_wor_perfplus=max_wor_perfplus, max_gor=max_gor,
                         max_gor_plug=max_gor_plug, max_gor_plug_plus=max_gor_plug_plus, max_gor_perf=max_gor_perf,
                         max_gor_perfplus=max_gor_perfplus, max_lgr=max_lgr, max_lgr_plug=max_lgr_plug,
                         max_lgr_plug_plus=max_lgr_plug_plus, max_lgr_perf=max_lgr_perf,
                         max_lgr_perfplus=max_lgr_perfplus,
                         convert_qmult_to_reservoir_barrels=convert_qmult_to_reservoir_barrels, active_node=active_node,
                         gor_limit=gor_limit,
                         )

        if date is None and properties_dict is not None:
            if 'date' not in properties_dict or not isinstance(properties_dict['date'], str):
                raise ValueError(f"No valid Date found for object with properties: {properties_dict}")
            self.date = properties_dict['date']
        else:
            self.date = date

        if properties_dict is None:
            return

        # Set the date related properties, then set the date, automatically setting the ISODate
        protected_attributes = ['date_format', 'start_date', 'unit_system']

        for attribute in protected_attributes:
            if attribute in properties_dict:
                self.__setattr__(f"_{attribute}", properties_dict[attribute])

        # Loop through the properties dict if one is provided and set those attributes
        remaining_properties = [x for x in properties_dict.keys() if x not in protected_attributes]
        for key in remaining_properties:
            self.__setattr__(key, properties_dict[key])

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        nexus_mapping: dict[str, tuple[str, type]] = {
            'WELL': ('well_name', str),
            'NAME': ('name', str),
            # Special keywords - QMULT
            'QALLRMAX_MULT': ('convert_qmult_to_reservoir_barrels', bool),
            'QOSMAX_MULT': ('use_qmult_qoil_surface_rate', bool),
            'QWSMAX_MULT': ('use_qmult_qwater_surface_rate', bool),
            'QGSMAX_MULT': ('use_qmult_qgas_surface_rate', bool),
            'QLIQSMAX_MULT': ('use_qmult_qoilqwat_surface_rate', bool),
            # Special Clear keywords
            'CLEAR': ('clear_all', bool),
            'CLEARQ': ('clear_q', bool),
            'CLEARP': ('clear_p', bool),
            'CLEARLIMIT': ('clear_limit', bool),
            'CLEARALQ': ('clear_alq', bool),
        }
        nexus_mapping.update(NexusConstraint.get_limit_constraints_map())
        nexus_mapping.update(NexusConstraint.get_pressure_constraints_map())
        nexus_mapping.update(NexusConstraint.get_rate_constraints_map())
        nexus_mapping.update(NexusConstraint.get_alq_constraints_map())
        return nexus_mapping

    @staticmethod
    def get_rate_constraints_map() -> dict[str, tuple[str, type]]:
        """Returns mapping of rate constraints."""

        nexus_mapping: dict[str, tuple[str, type]] = {
            'QALLRMAX': ('max_qmult_total_reservoir_rate', float),

            'QOSMAX': ('max_surface_oil_rate', float),
            'QGSMAX': ('max_surface_gas_rate', float),
            'QWSMAX': ('max_surface_water_rate', float),
            'QLIQSMAX': ('max_surface_liquid_rate', float),
            'QMHCMAX': ('max_hc_molar_rate', float),
            'QOSMAX-': ('max_reverse_surface_oil_rate', float),
            'QGSMAX-': ('max_reverse_surface_gas_rate', float),
            'QWSMAX-': ('max_reverse_surface_water_rate', float),
            'QLIQSMAX-': ('max_reverse_surface_liquid_rate', float),
            'QOMAX': ('max_reservoir_oil_rate', float),
            'QGMAX': ('max_reservoir_gas_rate', float),
            'QWMAX': ('max_reservoir_water_rate', float),
            'QLIQMAX': ('max_reservoir_liquid_rate', float),
            'QALLMAX': ('max_reservoir_total_fluids_rate', float),
            'QHCMAX': ('max_reservoir_hc_rate', float),
            'QOMAX-': ('max_reverse_reservoir_oil_rate', float),
            'QGMAX-': ('max_reverse_reservoir_gas_rate', float),
            'QWMAX-': ('max_reverse_reservoir_water_rate', float),
            'QLIQMAX-': ('max_reverse_reservoir_liquid_rate', float),
            'QALLMAX-': ('max_reverse_reservoir_total_fluids_rate', float),
            'QHCMAX-': ('max_reverse_reservoir_hc_rate', float),
            'QOSMIN': ('min_surface_oil_rate', float),
            'QGSMIN': ('min_surface_gas_rate', float),
            'QWSMIN': ('min_surface_water_rate', float),
            'QLIQSMIN': ('min_surface_liquid_rate', float),
            'QOMIN': ('min_reservoir_oil_rate', float),
            'QGMIN': ('min_reservoir_gas_rate', float),
            'QWMIN': ('min_reservoir_water_rate', float),
            'QLIQMIN': ('min_reservoir_liquid_rate', float),
            'QALLMIN': ('min_reservoir_total_fluids_rate', float),
            'QHCMIN': ('min_reservoir_hc_rate', float),
            'QOSMIN-': ('min_reverse_surface_oil_rate', float),
            'QGSMIN-': ('min_reverse_surface_gas_rate', float),
            'QWSMIN-': ('min_reverse_surface_water_rate', float),
            'QLIQSMIN-': ('min_reverse_surface_liquid_rate', float),
            'QOMIN-': ('min_reverse_reservoir_oil_rate', float),
            'QGMIN-': ('min_reverse_reservoir_gas_rate', float),
            'QWMIN-': ('min_reverse_reservoir_water_rate', float),
            'QLIQMIN-': ('min_reverse_reservoir_liquid_rate', float),
            'QALLMIN-': ('min_reverse_reservoir_total_fluids_rate', float),
            'QHCMIN-': ('min_reverse_reservoir_hc_rate', float),

            'QOIL': ('qmult_oil_rate', float),
            'QWATER': ('qmult_water_rate', float),
            'QGAS': ('qmult_gas_rate', float),
            'DPBHAVG': ('max_avg_comp_dp', float),
            'DPBHMX': ('max_comp_dp', float),
            'DPBHMX_ZONE': ('max_comp_dp_by_zone', dict[int, float]),
        }
        return nexus_mapping

    @staticmethod
    def get_pressure_constraints_map() -> dict[str, tuple[str, type]]:
        """Returns mapping of pressure constraints."""
        nexus_mapping: dict[str, tuple[str, type]] = {
            'PMIN': ('min_pressure', float),
            'PMAX': ('max_pressure', float),
            'PWMAX': ('max_wag_water_pressure', float),
            'PGMAX': ('max_wag_gas_pressure', float),
            'BHP': ('bottom_hole_pressure', float),
            'THP': ('tubing_head_pressure', float),
        }
        return nexus_mapping

    @staticmethod
    def get_limit_constraints_map() -> dict[str, tuple[str, type]]:
        """Returns a mapping of limit constraints."""
        nexus_mapping: dict[str, tuple[str, type]] = {
            'WCUTMAX': ('max_watercut', float),
            'WCUTPLUG': ('max_watercut_plug', float),
            'WCUTPLUGPLUS': ('max_watercut_plugplus', float),
            'WCUTPERF': ('max_watercut_perf', float),
            'WCUTPERFPLUS': ('max_watercut_perfplus', float),
            'WORMAX': ('max_wor', float),
            'WORPLUG': ('max_wor_plug', float),
            'WORPLUGPLUS': ('max_wor_plug_plus', float),
            'WORPERF': ('max_wor_perf', float),
            'WORPERFPLUS': ('max_wor_perfplus', float),
            'GORLIM': ('gor_limit', float),
            'EXPONENT': ('gor_limit_exponent', float),
            'FREQUENCY': ('gor_limit_frequency', float),
            'GORMAX': ('max_gor', float),
            'GORPLUG': ('max_gor_plug', float),
            'GORPLUGPLUS': ('max_gor_plug_plus', float),
            'GORPERF': ('max_gor_perf', float),
            'GORPERFPLUS': ('max_gor_perfplus', float),
            'LGRMAX': ('max_lgr', float),
            'LGRPLUG': ('max_lgr_plug', float),
            'LGRPLUGPLUS': ('max_lgr_plug_plus', float),
            'LGRPERF': ('max_lgr_perf', float),
            'LGRPERFPLUS': ('max_lgr_perfplus', float),
            'CGLIM': ('max_cum_gas_prod', float),
            'CWLIM': ('max_cum_water_prod', float),
            'COLIM': ('max_cum_oil_prod', float),
        }
        return nexus_mapping

    @staticmethod
    def get_alq_constraints_map() -> dict[str, tuple[str, type]]:
        """Gets the nexus mapping for artificial lift constraints."""
        nexus_mapping: dict[str, tuple[str, type]] = {
            'ALQ': ('artificial_lift_number', int),
            'SETTING': ('max_choke_setting', float),
            'GLEFMIN': ('min_gas_lift_efficiency', float),
            'GLRADD': ('gl_additive_correction', float),
            'ACTIVATE': ('active_node', bool),
            'POWER': ('pump_power', float),
            'SPEED': ('pump_speed', float),
            'CHOKELIMIT': ('choke_limit', str),
            'POSITION': ('manifold_position', int),
        }
        return nexus_mapping

    def update(self, new_data: dict[str, None | int | str | float | UnitSystem | dict],
               nones_overwrite: bool = False) -> None:
        """Updates attributes in the object based on the dictionary provided."""
        protected_attributes = ['date', 'date_format', 'start_date', 'unit_system']
        for key, value in new_data.items():
            modified_key = key
            if key in protected_attributes:
                modified_key = '_' + key
            if key == 'max_comp_dp_by_zone':
                # Special case for max_comp_dp_by_zone which is a dict[int, float]
                if isinstance(value, dict):
                    getattr(self, 'max_comp_dp_by_zone', {}).update(value)
            elif value is not None or nones_overwrite:
                setattr(self, modified_key, value)

    def to_table_line(self, headers: list[str]) -> str:
        """String representation of the constraint for entry to an inline constraint table.

        Args:
            headers (list[str]): Unused for nexusconstraint, provide an empty list
        """
        qmult_control_key_words = ['QALLRMAX_MULT', 'QOSMAX_MULT', 'QWSMAX_MULT', 'QGSMAX_MULT', 'QLIQSMAX_MULT']
        skip_attributes = ['date', 'unit_system', 'NAME', 'ACTIVATE', 'QOIL', 'QWATER', 'QGAS', 'WELL', 'control_mode']
        clear_attributes = ['CLEAR', 'CLEARQ', 'CLEARP', 'CLEARLIMIT', 'CLEARALQ']

        if headers:
            warnings.warn('Headers is currently not used in the constraint to line call')

        if self.name is not None:
            constraint_string = self.name
        elif self.well_name is not None:
            constraint_string = self.well_name
        else:
            raise ValueError('Must have a well or node name for returning a constraint to a string')

        for attribute, value in self.to_dict(keys_in_keyword_style=True).items():
            if value and attribute in qmult_control_key_words:
                constraint_string += (' ' + attribute.replace('_MULT', '') + ' MULT')
            elif value is None or attribute in skip_attributes:
                continue
            elif value and attribute in clear_attributes:
                constraint_string += ' ' + attribute
            elif attribute == 'DPBHMX_ZONE' and isinstance(value, dict):
                # Special case for DPBHMX_ZONE which is a dict[int, float]
                for zone, zone_value in value.items():
                    constraint_string += f' {attribute} {zone} {zone_value}'
            else:
                constraint_string += (' ' + attribute + ' ' + str(value))

        if self.active_node:
            constraint_string += ' ACTIVATE'
        elif self.active_node is not None:
            # equivalent to active node being False
            constraint_string += ' DEACTIVATE'

        constraint_string += '\n'
        return constraint_string

    def write_qmult_table(self) -> list[str]:
        """Writes out a QMULT table from a constraint that uses the following attributes.

        'QOIL': ('qmult_oil_rate', float).
        'QWATER': ('qmult_water_rate', float).
        'QGAS': ('qmult_gas_rate', float).

        Returns:
            list[str] with a representation of the QMULT table with a new line as a new entry in the list.
        """
        table_to_return = ['QMULT\n', 'WELL QOIL QGAS QWATER\n']
        qmult_values = self.write_qmult_values()
        table_to_return.append(qmult_values)
        table_to_return.append('ENDQMULT\n')

        return table_to_return

    def write_qmult_values(self) -> str:
        """Writes out the values for a QMULT table, callable on its own or using the write_qmult_table method."""
        qmult_values_keywords = ['qmult_oil_rate', 'qmult_gas_rate', 'qmult_water_rate']
        if self.name is not None:
            string_to_return = self.name
        elif self.well_name is not None:
            string_to_return = self.well_name
        else:
            raise ValueError('Must have a well or node name for returning a qmult table')

        for keyword in qmult_values_keywords:
            value = getattr(self, keyword, None)
            if value is None:
                string_to_return += ' NA'
            else:
                string_to_return += f' {str(value)}'
        string_to_return += '\n'
        return string_to_return



"""Class for collection of Nexus separator property methods."""
from __future__ import annotations
from dataclasses import dataclass, field
import os
from typing import Optional, MutableMapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.FcsFile import FcsNexusFile
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusSeparatorMethod import NexusSeparatorMethod
from ResSimpy.DataModelBaseClasses.Separator import Separator
from ResSimpy.Utils.dynamic_method_manipulations import add_dynamic_method


@dataclass(kw_only=True)
class NexusSeparatorMethods(Separator):
    """Class for collection of Nexus separator property methods.

    Attributes:
        inputs (dict[int, NexusSeparatorMethod]): Dictionary collection of Nexus separator property methods
        files (dict[int, NexusFile]): Dictionary collection of separator property files, defined in Nexus fcs.
    """

    __inputs: MutableMapping[int, NexusSeparatorMethod]
    __files: dict[int, NexusFile]
    __properties_loaded: bool = False  # Used in lazy loading
    __model_unit_system: UnitSystem
    _model_files: Optional[FcsNexusFile] = field(default=None, repr=False, compare=False)

    def __init__(self, model_unit_system: UnitSystem,
                 inputs: Optional[MutableMapping[int, NexusSeparatorMethod]] = None,
                 files: Optional[dict[int, NexusFile]] = None, assume_loaded: bool = False,
                 model_files: Optional[FcsNexusFile] = None) -> None:
        """Initialises the NexusSeparatorMethods class.

        Args:
            model_unit_system (UnitSystem): Unit system used in the model.
            inputs (Optional[MutableMapping[int, NexusSeparatorMethod]]): Collection of Nexus separator property
            methods.
            files (Optional[dict[int, NexusFile]]): Collection of separator property files, as defined in Nexus fcs
            file. Keyed by the method number.
            assume_loaded (bool): If True, assumes that the separator methods have already been loaded.
            model_files (Optional[FcsNexusFile]): The Nexus fcs file that contains the whole set of model files.
        """
        if inputs:
            self.__inputs = inputs
        else:
            self.__inputs: MutableMapping[int, NexusSeparatorMethod] = {}
        if files:
            self.__files = files
        else:
            self.__files = {}
        self.__model_unit_system = model_unit_system
        self._model_files = model_files
        self.__properties_loaded = assume_loaded
        super().__init__()

    def __repr__(self) -> str:
        """Pretty printing separator methods."""
        if not self.__properties_loaded:
            self.load_separator_methods()
        printable_str = ''
        for table_num in self.__inputs.keys():
            printable_str += '\n--------------------------------\n'
            printable_str += f'SEPARATOR method {table_num}\n'
            printable_str += '--------------------------------\n'
            sep_type = self.__inputs[table_num].separator_type
            printable_str += f"\nSEPARATOR_TYPE: {'' if sep_type is None else sep_type.value}\n"
            printable_str += self.__inputs[table_num].__repr__()
            printable_str += '\n'

        return printable_str

    @property
    def inputs(self) -> MutableMapping[int, NexusSeparatorMethod]:
        """Returns mapping of Nexus separator methods."""
        if not self.__properties_loaded:
            self.load_separator_methods()
        return self.__inputs

    @property
    def files(self) -> dict[int, NexusFile]:
        """Returns dictionary of nexus files.

        Returns:
            NexusFile(dict): Nexus files where keys are of type int.
        """
        return self.__files

    def load_separator_methods(self) -> None:
        """Loads and processes the seperator method files."""
        # Read in separator properties from Nexus separator method files
        if self.__files is not None and len(self.__files) > 0:  # Check if separator files exist
            for table_num in self.__files.keys():  # For each separator property method
                separator_file = self.__files[table_num]
                if separator_file.location is None:
                    raise ValueError(f'Unable to find separator file: {separator_file}')
                if os.path.isfile(separator_file.location):
                    # Create NexusSeparatorMethod object
                    self.__inputs[table_num] = NexusSeparatorMethod(file=separator_file, input_number=table_num,
                                                                    model_unit_system=self.__model_unit_system)
                    # Populate object with separator properties in input file
                    self.__inputs[table_num].read_properties()
        self.__properties_loaded = True

    @property
    def model_unit_system(self) -> UnitSystem:
        """Return the model unit system."""
        return self.__model_unit_system

    @property
    def keyword(self) -> str:
        """Return the keyword for separator methods."""
        return 'SEPARATOR'

    def add_method(self, method: DynamicProperty, new_file_name: str, create_new_file: bool = False) -> None:
        """Adds a new separator method to the collection.

        Args:
            method (DynamicProperty): The separator method to add.
            new_file_name (str): The name of the file to save the method to.
            create_new_file (bool): Whether to create a new file for the method.
        """
        add_dynamic_method(dynamic_method_collection=self, method=method, new_file_name=new_file_name,
                           create_new_file=create_new_file)

    def _method_type(self) -> type[NexusSeparatorMethod]:
        """Returns the expected type of the dynamic property."""
        return NexusSeparatorMethod



"""Grid base class."""

from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass

from typing import Optional, Sequence

import numpy as np

from ResSimpy.DataModelBaseClasses.GridArrayDefinition import GridArrayDefinition
from ResSimpy.DataModelBaseClasses.GridArrayFunction import GridArrayFunction
from ResSimpy.DataModelBaseClasses.Over import Over
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.FileOperations.File import File
from ResSimpy.GenericContainerClasses.LGRs import LGRs


@dataclass(kw_only=True)
class Grid(ABC):
    """A base class to represent a collection of grids in a reservoir simulation model."""
    # Grid dimensions
    _range_x: Optional[int]
    _range_y: Optional[int]
    _range_z: Optional[int]

    _netgrs: GridArrayDefinition
    _porosity: GridArrayDefinition
    _sw: GridArrayDefinition
    _sg: GridArrayDefinition
    _pressure: GridArrayDefinition
    _temperature: GridArrayDefinition
    _kx: GridArrayDefinition
    _ky: GridArrayDefinition
    _kz: GridArrayDefinition
    _iregion: dict[str, GridArrayDefinition]
    _grid_properties_loaded: bool = False
    _lgrs: LGRs
    _overs: Sequence[Over]

    def __init__(self, assume_loaded: bool = False) -> None:
        """Initialises the Grid class."""
        self._netgrs = GridArrayDefinition()
        self._porosity = GridArrayDefinition()
        self._sw = GridArrayDefinition()
        self._sg = GridArrayDefinition()
        self._pressure = GridArrayDefinition()
        self._temperature = GridArrayDefinition()
        self._kx = GridArrayDefinition()
        self._ky = GridArrayDefinition()
        self._kz = GridArrayDefinition()
        self._iregion = {}

        # Grid dimensions
        self._range_x: Optional[int] = None
        self._range_y: Optional[int] = None
        self._range_z: Optional[int] = None
        self._grid_properties_loaded = assume_loaded

        # LGRs
        self._lgrs: LGRs = LGRs()

    @property
    def range_x(self) -> int | None:
        """Returns an instance of grid dimensions.
        Loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._range_x

    @property
    def range_y(self) -> int | None:
        """Returns an instance of grid dimensions.
        Loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._range_y

    @property
    def range_z(self) -> int | None:
        """Returns range_z.
        Loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._range_z

    @property
    def netgrs(self) -> GridArrayDefinition:
        """Returns grid array definition for netgrs.
        Loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._netgrs

    @property
    def porosity(self) -> GridArrayDefinition:
        """Returns grid porosity.
        Loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._porosity

    @property
    def sw(self) -> GridArrayDefinition:
        """Returns water saturation grid.
        Loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._sw

    @property
    def sg(self) -> GridArrayDefinition:
        """Returns gas saturation grid.
        Loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._sg

    @property
    def pressure(self) -> GridArrayDefinition:
        """Returns grid pressure.
        Loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._pressure

    @property
    def temperature(self) -> GridArrayDefinition:
        """Returns grid temperature.
        Loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._temperature

    @property
    def kx(self) -> GridArrayDefinition:
        """Returns grid kx.
        Loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._kx

    @property
    def ky(self) -> GridArrayDefinition:
        """Returns grid ky.
        Loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._ky

    @property
    def kz(self) -> GridArrayDefinition:
        """Returns grid kz.
        loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._kz

    @property
    def iregion(self) -> dict[str, GridArrayDefinition]:
        """Returns integer regions as dictionary keyed by the name of the array.
        loads grid properties if not loaded.
        """
        self.load_grid_properties_if_not_loaded()
        return self._iregion

    @property
    def lgrs(self) -> LGRs:
        """Returns the LGR module object associated with the grid."""
        return self._lgrs

    @abstractmethod
    def load_structured_grid_file(self, structure_grid_file: File,
                                  model_unit_system: UnitSystem, lazy_loading: bool) -> Grid:
        """Loads in a structured grid file with all grid properties, and the array functions defined with 'FUNCTION'.

        Args:
            structure_grid_file(File, lazy_loading: bool):The NexusFile representation of a structured grid file for
            converting into a structured grid file class.
            lazy_loading(bool): If set to True, parts of the grid will only be loaded in
             when requested via properties on the object.
            model_unit_system(UnitSystem): The default unit system used in the model.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def load_grid_properties_if_not_loaded(self) -> None:
        """Loads grid properties if not loaded."""
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def to_dict(self) -> dict[str, Optional[int] | GridArrayDefinition]:
        """Converts object to a dictionary."""
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def update_properties_from_dict(self, data: dict[str, int | GridArrayDefinition]) -> None:
        """Allows you to update properties on the class using the provided dict of values.

        Args:
            data(dict[str, int | GridArrayDefinition]): Dictionary of values to update on the class.
        """
        raise NotImplementedError("Implement this in the derived class")

    @property
    def array_functions(self) -> Optional[Sequence[GridArrayFunction]]:
        """Returns a list of the array functions defined in the structured grid file."""
        raise NotImplementedError("Implement this in the derived class")

    def grid_array_definition_to_numpy_array(self, grid_array_definition: GridArrayDefinition) -> np.ndarray:
        """Converts a grid array to a numpy array."""
        return grid_array_definition.get_array_from_file(self.range_x, self.range_y, self.range_z)

    @property
    def overs(self) -> Sequence[Over]:
        """Returns the overs associated with the grid."""
        return self._overs



"""Class that holds all the constraints for a NexusNetwork object.

This class is responsible for loading, modifying and removing constraints from the NexusNetwork object. This class is
also responsible for writing the constraints to the Nexus deck. An instance of this is held within the Network class
as "constraints".
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional, cast
from uuid import UUID

import pandas as pd
import re

from ResSimpy.DataModelBaseClasses.Constraint import Constraint
from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.GenericContainerClasses.Constraints import Constraints
from ResSimpy.Nexus.nexus_collect_tables import collect_all_tables_to_objects
from ResSimpy.Nexus.DataModels.Network.NexusConstraint import NexusConstraint
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Time.ISODateTime import ISODateTime
from ResSimpy.Utils.obj_to_dataframe import obj_to_dataframe
import ResSimpy.Nexus.nexus_file_operations as nfo

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator


@dataclass
class NexusConstraints(Constraints):
    """Class that holds all the constraints for a NexusNetwork object.

    This class is responsible for loading, modifying and removing constraints from the NexusNetwork object. This
    class is also responsible for writing the constraints to the Nexus deck.

    Attributes:
        constraints: dict[str, list[NexusConstraint]]: dictionary of all constraints defined within a model,
        keyed by the node or well name.
    """
    _constraints: dict[str, list[NexusConstraint]] = field(default_factory=dict)

    def __init__(self, parent_network: NexusNetwork, model: NexusSimulator) -> None:
        """Initialises the NexusConstraints class.

        Args:
            parent_network (NexusNetwork): the network object that the constraints are associated with.
            model (NexusSimulator): the model object that the constraints are associated with.
        """
        self.__parent_network: NexusNetwork = parent_network
        self._constraints: dict[str, list[NexusConstraint]] = {}
        self.__model: NexusSimulator = model

    def get_all(self, object_name: Optional[str] = None, date: Optional[str] = None) -> \
            dict[str, list[NexusConstraint]]:
        """Get the constraints of the existing model with optional parameters to filter for name and date.

        Args:
            object_name (Optional[str]): name of the connection, node or wellname to return. Defaults to None.
            date (Optional[str]): date in model format to filter the dates to in the constraints
        Returns: dict[str, list[NexusConstraint]] dictionary of all constraints defined within a model, keyed by the \
            name of the well/node.
        """
        self.__parent_network.get_load_status()

        if object_name is None:
            constraints_to_return = self._constraints
        else:
            constraints_to_return = {k: v for k, v in self._constraints.items() if k == object_name}

        if date is None:
            return constraints_to_return

        date_filtered_constraints = {}
        for constraint_name, constraint_list in constraints_to_return.items():
            new_constraint_list = [x for x in constraint_list if x.date == date]
            if len(new_constraint_list) > 0:
                date_filtered_constraints[constraint_name] = new_constraint_list
        return date_filtered_constraints

    def get_df(self) -> pd.DataFrame:
        """Creates a dataframe representing all processed constraint data in a surface file.

        Returns:
            DataFrame: of the properties of the constraint through time with each row representing \
                a change in constraint.
        """
        self.__parent_network.get_load_status()
        list_constraints: list[Constraint] = []
        for well_constraints in self._constraints.values():
            list_constraints.extend(well_constraints)
        obj_to_dataframe(list_constraints)

        return obj_to_dataframe(list_constraints)

    def get_overview(self) -> str:
        """Returns overview of Constraints."""
        raise NotImplementedError('To be implemented')

    def load(self, surface_file: NexusFile, start_date: str, default_units: UnitSystem) -> None:
        """Loads nexus surface file.

        Args:
            surface_file(NexusFile): File path of the Nexus surface file.
            start_date(str): Start date of the run.
            default_units(UnitSystem): default units used if not units are found.
        """
        # CONSTRAINT keyword represents a table with a header and columns.
        # CONSTRAINTS keyword represents a list of semi structured constraints with a well_name and then constraints
        if surface_file.file_content_as_list is not None:
            contents_str = "\n".join(surface_file.file_content_as_list)
            contents_str_line_continuation = re.sub(">\\s*\\n", "", contents_str)
            new_contents_list = contents_str_line_continuation.splitlines()
            surface_file.file_content_as_list = new_contents_list
        _, new_constraints = collect_all_tables_to_objects(surface_file,
                                                           {
                                                               'CONSTRAINTS': NexusConstraint,
                                                               'CONSTRAINT': NexusConstraint,
                                                               'QMULT': NexusConstraint
                                                           },
                                                           start_date=start_date,
                                                           default_units=default_units,
                                                           date_format=self.__model.date_format)
        self._add_to_memory(new_constraints)

    def _add_to_memory(self, additional_constraints: Optional[dict[str, list[NexusConstraint]]]) -> None:
        """Adds additional constraints to memory within the NexusConstraints object.

            If user adds constraints list this will not be reflected in the Nexus deck at this time.

        Args:
        ----
            additional_constraints (Optional[dict[str, list[NexusConstraint]]]): additional constraints to add as a list
        """
        if additional_constraints is None:
            return
        for name, constraints in additional_constraints.items():
            # this does not guarantee the constraints are ordered by datetime
            existing_constraints = self._constraints.get(name, [])
            existing_constraints.extend(constraints)
            self._constraints[name] = existing_constraints

    def find_by_properties(self, object_name: str, constraint_dict: DataObjectMixinDictType) -> \
            NexusConstraint:
        """Finds a uniquely matching constraint from a given set of properties in a dictionary of attributes.

        Args:
            object_name (str): name of the node to which the constraint is applied (node name/well name)
            constraint_dict (dict[str, float | str | int]): dictionary of attributes to match on. \
            Allows for partial matches if it finds a unique constraint.

        Returns:
            NexusConstraint of an existing constraint in the model that uniquely matches the provided \
            constraint_dict constraint
        """
        self.__parent_network.get_load_status()
        found_object_from_network = self.__parent_network.find_network_element_with_dict(object_name, constraint_dict,
                                                                                         'constraints')
        # ensure typing is consistent
        if isinstance(found_object_from_network, NexusConstraint):
            return found_object_from_network
        else:
            raise TypeError(f'Wrong object type returned expected NexusConstraint, '
                            f'instead returned {type(found_object_from_network)}')

    def remove(self, constraint_dict: Optional[DataObjectMixinDictType] = None,
               constraint_id: Optional[UUID] = None) -> None:
        """Remove a constraint based on closest matching constraint, requires node name and date.

        Needs one of at least constraint dict or constraint id.

        Args:
            constraint_dict (Optional[dict[str, float | str | int]]): Constraint matching these attributes will be
                removed. Defaults to None.
            constraint_id (Optional[UUID]): Constraint matching this id will be removed.
                Will not be used if constraint dict is provided. Defaults to None.
        """
        self.__parent_network.get_load_status()

        if constraint_dict is None and constraint_id is None:
            raise ValueError('no options provided for both constraint_id and constraint_dict')
        if constraint_dict is not None:
            name = constraint_dict.get('name', None)
            if name is None:
                raise ValueError(f'No well or node name provided instead got {name}')
            name = str(name)
            # check for wildcards
            if '*' in name:
                raise NotImplementedError(f'Removing constraints with wildcards is currently unsupported, for {name=}')
            constraint_to_remove = self.find_by_properties(name, constraint_dict)
            constraint_id = constraint_to_remove.id
        if constraint_id is None:
            raise ValueError(f'No constraint found with {constraint_id=}')
        # find which file and remove from the file as list
        surface_file = self.__find_which_surface_file_from_id(constraint_id)
        surface_file.remove_object_from_file_as_list([constraint_id])
        # remove from memory
        removed_constraint = None
        for name, list_constraints in self._constraints.items():
            for i, constraint in enumerate(list_constraints):
                if constraint.id == constraint_id:
                    removed_constraint = list_constraints.pop(i)
        # remove the constraints table if there are no more for that date timestamp
        if removed_constraint is None:
            raise ValueError(f'No constraint found with {constraint_id=}')
        date_of_constraint_removed = removed_constraint.date
        if date_of_constraint_removed is None:
            return
        # search the loaded constraints for constraints on the same date as the removed constraint
        # if there are none then remove the table
        if not any(z for x, y in self._constraints.items() for z in y if z.date == date_of_constraint_removed):
            self.__remove_empty_token_table(date_of_constraint_removed, surface_file=surface_file, token='CONSTRAINTS')
            self.__remove_empty_token_table(date_of_constraint_removed, surface_file=surface_file, token='QMULT',
                                            remove_header=True)
        return

    def __remove_empty_token_table(self, date: str, surface_file: NexusFile, token: str = 'CONSTRAINTS',
                                   remove_header: bool = False) -> None:
        """Removes a table for a given date if it has no other dates for that constraint."""
        # TODO promote this to a higher level function
        keyword_map = {x: y[0] for x, y in NexusConstraint.get_keyword_mapping().items()}
        file_content = surface_file.get_flat_list_str_file
        token_line_number = -1
        end_token_line_number = -1
        found_date = False
        for i, line in enumerate(file_content):
            if nfo.check_token('TIME', line) and \
                    nfo.get_expected_token_value(token='TIME', token_line=line, file_list=file_content[i:]) == date:
                found_date = True
            if found_date and nfo.check_token(token, line):
                token_line_number = i
            elif found_date and nfo.check_token('END' + token, line):
                end_token_line_number = i
            if token_line_number > 0 and end_token_line_number > 0:
                break
        if end_token_line_number > token_line_number:
            if remove_header:
                header_index, headers = nfo.get_table_header(file_content[token_line_number:end_token_line_number],
                                                             keyword_map)
                header_index = token_line_number + header_index
            # remove the last one first to avoid index errors
            surface_file.remove_from_file_as_list(end_token_line_number)
            if remove_header:
                surface_file.remove_from_file_as_list(header_index)
            surface_file.remove_from_file_as_list(token_line_number)

    def __find_which_surface_file_from_id(self, constraint_id: UUID) -> NexusFile:
        """Finds the surface file with the object id requested."""
        # TODO: make this generic with the find_which_wellspec_file_from_completion_id.

        if self.__model.model_files.surface_files is None:
            raise ValueError(f'No surface file found in fcs file at {self.__model.model_files.location}')
        surface_files = [x for x in self.__model.model_files.surface_files.values() if
                         x.object_locations is not None and constraint_id in x.object_locations]
        if len(surface_files) == 0:
            raise FileNotFoundError(f'No surface file found with an existing constraint that has: {constraint_id=}')
        surface_file = surface_files[0]
        if surface_file.file_content_as_list is None:
            raise FileNotFoundError(f'No file content found in file: {surface_file.location} '
                                    f'with an existing constraint that has: {constraint_id=}')
        return surface_file

    def add(self, constraint_to_add: DataObjectMixinDictType | Constraint,
            comments: str | None = None) -> None:
        """Adds a constraint to the network and corresponding surface file.

        Args:
            constraint_to_add (dict[str, float | int | str | UnitSystem] | NexusConstraint): properties of \
            the constraints or a constraint object
            comments (str | None): add optional post line comments
        """
        self.__parent_network.get_load_status()

        # add to memory
        if isinstance(constraint_to_add, dict):
            name = constraint_to_add.get('name', None)
            if name is None:
                raise ValueError('Input arguments or constraint_to_add dictionary must contain a name for the node.')
            new_constraint = NexusConstraint(constraint_to_add)
        else:
            new_constraint = cast(NexusConstraint, constraint_to_add)
            name = new_constraint.name
        if name is None:
            raise ValueError('No name found in the provided constraint object.')
        # check for wildcards
        name = str(name)
        if '*' in name:
            raise NotImplementedError('Adding constraints with wildcards is currently unsupported')
        self._add_to_memory({name: [new_constraint]})

        # add to the file
        if self.__model.model_files.surface_files is None:
            raise FileNotFoundError('No well file found, cannot modify ')

        file_to_add_to = self.__model.model_files.surface_files[1]

        file_as_list = file_to_add_to.get_flat_list_str_file
        if file_as_list is None:
            raise ValueError(f'No file content found in the surface file specified at {file_to_add_to.location}')

        constraint_date = new_constraint.date
        if constraint_date is None:
            raise ValueError(f'Require date for adding constraint to, instead got {new_constraint.date}')
        new_constraint_text = []
        date_comparison = -1
        date_index = -1
        new_constraint_index = -1
        id_line_locs = []
        new_table_needed = False
        new_date_needed = False
        new_qmults_table_needed = False
        correct_for_padding = 0
        # check for need to add qmult table
        qmult_keywords = ['qmult_oil_rate', 'qmult_gas_rate', 'qmult_water_rate']
        # if any of the qmults are defined in the new constraint then add a qmult table
        add_qmults = any(getattr(new_constraint, x, None) for x in qmult_keywords)
        if len(file_as_list) < 2:
            # pad the file if it is empty or only has one line
            new_table_needed = True
            file_as_list.append('\n')
            # ensure the padding is added to the file through the new_constraint_text at the end of the file.
            new_constraint_text.append('\n')
            correct_for_padding = -1
        for index, line in enumerate(file_as_list):
            if nfo.check_token('TIME', line):
                constraint_date_from_file = nfo.get_expected_token_value('TIME', line, [line])
                date_comparison = self.__model._sim_controls.compare_dates(constraint_date_from_file, constraint_date)
                if date_comparison == 0:
                    date_index = index
                    continue

                elif date_comparison > 0 and date_index >= 0:
                    # if a date that is greater than the additional constraint then we have overshot and need to
                    # add in a new table or time card
                    # this is the case where we don't need to write a new time card
                    new_table_needed = True
                    new_constraint_index = index - 1

                elif date_comparison > 0:
                    new_table_needed = True
                    new_date_needed = True
                    new_constraint_index = index
                else:
                    continue
            if nfo.check_token('ENDCONSTRAINTS', line) and date_comparison == 0:
                # find the end of a constraint table and add the new constraint
                new_constraint_index = index
                constraint_string = new_constraint.to_table_line([])
                new_constraint_text.append(constraint_string)
                id_line_locs = [new_constraint_index]
            elif index == len(file_as_list) - 1 and date_index < 0 and not nfo.check_token('ENDQMULT', line):
                # if we're on the final line of the file and we haven't yet set a constraint index
                new_table_needed = True
                new_constraint_index = index + 1
                new_date_needed = True
                if add_qmults:
                    new_qmults_table_needed = True
            elif index == len(file_as_list) - 1 and date_index == 0 and not nfo.check_token('ENDQMULT', line):
                new_table_needed = True
                new_constraint_index = index + 1
                new_date_needed = False
                if add_qmults:
                    new_qmults_table_needed = True

            if new_date_needed:
                # if the date card doesn't exist then add it to the file first
                new_constraint_text.append(f'TIME {constraint_date}\n')

            if new_table_needed:
                new_constraint_text.append('CONSTRAINTS\n')
                new_constraint_text.append(new_constraint.to_table_line([]))
                new_constraint_text.append('ENDCONSTRAINTS\n')
                new_constraint_text.append('\n')
                line_id = new_constraint_index + len(new_constraint_text) - 3 + correct_for_padding
                id_line_locs = [line_id]

            if add_qmults and new_qmults_table_needed:
                new_constraint_text.extend(new_constraint.write_qmult_table())
                # add id location for the qmult table as well
                line_id = new_constraint_index + len(new_constraint_text) - 2 + correct_for_padding
                id_line_locs.append(line_id)
                add_qmults = False
            elif add_qmults and nfo.check_token('ENDQMULT', line) and date_comparison == 0:
                # find the end of the table of qmults that already exist
                new_qmult_index = index
                qmult_string = new_constraint.write_qmult_values()
                new_qmult_object_ids = {new_constraint.id: [new_qmult_index]}
                file_to_add_to.add_to_file_as_list(additional_content=[qmult_string], index=new_qmult_index,
                                                   additional_objects=new_qmult_object_ids)
                add_qmults = False

            if new_constraint_index >= 0 and not add_qmults:
                # once we have found where to add constraint then add the constraint to file and update file ids
                new_constraint_object_ids = {
                    new_constraint.id: id_line_locs
                }
                file_to_add_to.add_to_file_as_list(additional_content=new_constraint_text, index=new_constraint_index,
                                                   additional_objects=new_constraint_object_ids, comments=comments)
                break

    def modify(self, name: str,
               current_constraint: DataObjectMixinDictType | Constraint,
               new_constraint_props: DataObjectMixinDictType | Constraint,
               comments: Optional[str] = None) \
            -> None:
        """Modify an existing constraint.

        Retains existing constraint values that are not overridden by the new constraint properties.

        Args:
            name (str): The well name
            current_constraint (dict[str, None | float | int | str] | Constraint): dictionary or constraint object\
            with enough attributes to identify a unique existing constraint in the model.
            new_constraint_props (dict[str, None | float | int | str] | Constraint): dictionary or constraint to \
            update the constraint with.
            comments (Optional[str]): comments to add inline to the constraint line in the surface file.
        """

        def clean_constraint_inputs(constraint: DataObjectMixinDictType | Constraint) -> DataObjectMixinDictType:
            """Cleans up an input ensuring consistent type is returned."""
            if isinstance(constraint, Constraint):
                cleaned_dict = constraint.to_dict()
            else:
                cleaned_dict = constraint
            return cleaned_dict

        cleaned_current_constraint = clean_constraint_inputs(current_constraint)
        cleaned_new_constraint = clean_constraint_inputs(new_constraint_props)

        existing_constraint_obj = self.find_by_properties(name, cleaned_current_constraint)
        self.remove(constraint_id=existing_constraint_obj.id)
        combination_of_constraints = existing_constraint_obj.to_dict()
        combination_of_constraints.update(cleaned_new_constraint)

        self.add(combination_of_constraints, comments)

    def to_string_for_date(self, date: ISODateTime) -> str:
        """Converts the constraints for a specific date to a string representation.

        Args:
            date (ISODateTime): The date for which to convert the constraints.

        Returns:
            str: String representation of the constraints for the specified date.
        """
        constraints_for_date = {name: [x for x in constraints if x.iso_date == date] for name, constraints in
                                self._constraints.items()}
        if not constraints_for_date:
            return ''

        printable_string = 'CONSTRAINTS\n'
        for name, constraints in constraints_for_date.items():
            for constraint in constraints:
                printable_string += constraint.to_table_line([])
        printable_string += 'ENDCONSTRAINTS\n'
        return printable_string



"""Class for collection of IPRs."""
from __future__ import annotations
from typing import Optional, MutableMapping, TYPE_CHECKING
import pandas as pd

import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusIPRMethod import NexusIprMethod
from ResSimpy.Nexus.nexus_file_operations import read_table_to_df

if TYPE_CHECKING:
    from ResSimpy import NexusSimulator


class NexusIprMethods:
    __inputs: MutableMapping[int, NexusIprMethod]
    __files: dict[int, NexusFile]
    __model: NexusSimulator
    __has_been_loaded: bool = False

    def __init__(self, model: NexusSimulator, tables: Optional[list[NexusIprMethod]] = None) -> None:
        """Class for collections of IPRs.

        Args:
            tables: Collections of IPR tables
            model: NexusSimulator instance
        """

        if tables is None:
            self.tables = []
        else:
            self.tables = tables

        self.__model = model

    def read_iprtables_as_df(self, file_as_list: list[str]) -> pd.DataFrame:
        """Reads in IPR files from Nexus into a dataframe.

        Args:
            file_as_list (list): File as list of strings.
        """

        date = ''
        reading_line = False
        table_lines: list[str] = []
        all_tables = []

        for line in file_as_list:
            if nfo.check_token('TIME', line):
                date = nfo.get_expected_token_value(token='TIME', token_line=line, file_list=[line])
            if nfo.check_token('IPRTABLE', line):
                reading_line = True
                continue
            if nfo.check_token('ENDIPRTABLE', line):
                reading_line = False

                df = read_table_to_df(file_as_list=table_lines)
                new_iprtable = NexusIprMethod(date=date, table=df)
                self.tables.append(new_iprtable)

                all_tables.append(df)
                table_lines = []
                continue

            if reading_line:
                table_lines.append(line)

        return pd.concat(all_tables, ignore_index=True) if all_tables else pd.DataFrame()

    def get_all(self) -> list[NexusIprMethod]:
        """Returns loaded IPRTables."""
        if not self.__has_been_loaded:
            self.load()
        return self.tables

    def load(self) -> None:
        """Loads IPRTables."""
        ipr_files = self.__model.model_files.ipr_files

        if ipr_files is not None:
            for ipr_file in ipr_files.values():
                self.read_iprtables_as_df(file_as_list=ipr_file.get_flat_list_str_file)
            self.__has_been_loaded = True
        else:
            raise ValueError("""IPR files are missing. Can not load IPRTables.""")



"""Contains a class for representing a set of WellList objects for the Nexus model."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal, Sequence

from ResSimpy.DataModelBaseClasses.NetworkList import NetworkList
from ResSimpy.GenericContainerClasses.NetworkLists import NetworkLists
from ResSimpy.Nexus.DataModels.NexusWellList import NexusWellList

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork

import pandas as pd


@dataclass(kw_only=True)
class NexusWellLists(NetworkLists):
    """Class for representing a set of WellList objects for the Nexus model."""

    _lists: list[NexusWellList] = field(default_factory=list)
    __parent_network: NexusNetwork

    def __init__(self, parent_network: NexusNetwork, well_lists: None | list[NexusWellList] = None) -> None:
        """Initialises the NexusWellLists class.

        Args:
            parent_network (NexusNetwork): The network that the well lists are a part of.
            well_lists (list[NexusWellList]): The well lists found in the network.
        """
        self._lists = well_lists if well_lists is not None else []
        self.__parent_network = parent_network
        super().__init__(parent_network)

    @property
    def _network_element_name(self) -> Literal['welllists']:
        return 'welllists'

    def get_all(self) -> list[NexusWellList]:
        """Returns all WellList names."""
        self.__parent_network.get_load_status()
        return self._lists

    def get_df(self) -> pd.DataFrame:
        """Creates a dataframe representing all processed node data in a surface file.

        Returns:
            DataFrame: of the properties of the nodes through time with each row representing a node.
        """
        self.__parent_network.get_load_status()
        df_store = pd.DataFrame()
        for welllist in self._lists:
            data = {'name': [welllist.name] * len(welllist.elements_in_the_list),
                    'well': welllist.elements_in_the_list,
                    'date': [welllist.date] * len(welllist.elements_in_the_list)}
            df_store = pd.concat([df_store, pd.DataFrame(data)], ignore_index=True)
        return df_store

    @property
    def welllists(self) -> list[NexusWellList]:
        """Returns list of all WellList objects."""
        self.__parent_network.get_load_status()
        return self._lists

    @staticmethod
    def table_header() -> Literal['WELLLIST']:
        """Start of the Node definition table."""
        return 'WELLLIST'

    @staticmethod
    def table_footer() -> Literal['ENDWELLLIST']:
        """End of the Node definition table."""
        return 'ENDWELLLIST'

    def get_all_by_name(self, well_list_name: str) -> list[NexusWellList]:
        """Returns a list of WellLists which match the provided name loaded from the simulator.

        Args:
        ----
            well_list_name (str): name of the requested WellList

        Returns:
        -------
            list[NexusWellList]: the requested list of filtered welllists with the same name
        """
        self.__parent_network.get_load_status()
        return [x for x in self._lists if x.name == well_list_name]

    def _add_to_memory(self, additional_list: None | Sequence[NetworkList]) -> None:
        """Extends the welllists stored in memory by a list of welllist provided to it."""
        if additional_list is None:
            return
        only_welllist_objects = [x for x in additional_list if isinstance(x, NexusWellList)]
        self._lists.extend(only_welllist_objects)



"""Enum representing Date Formats."""

from enum import Enum


class DateFormat(str, Enum):
    """Enum representing standard Nexus Date Formats (DATEFORMAT)."""
    DD_MM_YYYY = 'DD/MM/YYYY'  # e.g. 14/03/1999
    MM_DD_YYYY = 'MM/DD/YYYY'  # e.g. 03/14/1999
    DD_MMM_YYYY = 'MM DDD YYYY'  # e.g. 14 MAR 1999
    DD_MM_YYYY_h_m_s = 'DD/MM/YYYY(h:m:s)'  # e.g. 14/03/1999(09:43:18)
    MM_DD_YYYY_h_m_s = 'MM/DD/YYYY(h:m:s)'  # e.g. 03/14/1999(09:43:18)



from enum import Enum


class ActivationChangeEnum(str, Enum):
    """Enum representing whether an object has been activated or deactivated."""
    ACTIVATE = 'ACTIVATE'
    DEACTIVATE = 'DEACTIVATE'



"""Contains a class for representing a set of NodeList objects for the Nexus model."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal, Sequence

from ResSimpy.DataModelBaseClasses.NetworkList import NetworkList
from ResSimpy.GenericContainerClasses.NetworkLists import NetworkLists
from ResSimpy.Nexus.DataModels.Network.NexusNodeList import NexusNodeList

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


@dataclass(kw_only=True)
class NexusNodeLists(NetworkLists):
    """Class for representing a set of NodeList objects for the Nexus model."""

    _lists: list[NexusNodeList] = field(default_factory=list)
    __parent_network: NexusNetwork

    def __init__(
        self,
        parent_network: NexusNetwork,
        node_lists: None | list[NexusNodeList] = None,
    ) -> None:
        """Initialises the NexusNodeLists class. Supports CLEAR, ADD and REMOVE operations on the NodeList.

        Args:
            parent_network (NexusNetwork): The network that the node lists are a part of.
            node_lists (list[NexusNodeList]): The list of node lists to be added to the NexusNodeLists object.
        """
        self._lists = node_lists if node_lists is not None else []
        self.__parent_network = parent_network
        super().__init__(parent_network=parent_network)

    def get_all(self) -> list[NexusNodeList]:
        """Returns all NodeList names."""
        self.__parent_network.get_load_status()
        return self._lists

    @property
    def nodelists(self) -> list[NexusNodeList]:
        """Returns list of all NodeList objects."""
        self.__parent_network.get_load_status()
        return self._lists

    @staticmethod
    def table_header() -> Literal["NODELIST"]:
        """Start of the Node definition table."""
        return "NODELIST"

    @staticmethod
    def table_footer() -> Literal["ENDNODELIST"]:
        """End of the Node definition table."""
        return "ENDNODELIST"

    @property
    def _network_element_name(self) -> Literal["nodelists"]:
        return "nodelists"

    def _add_to_memory(self, additional_list: None | Sequence[NetworkList]) -> None:
        """Extends the nodelist stored in memory by a list of nodelist provided to it."""
        if additional_list is None:
            return
        if not isinstance(additional_list, list):
            additional_list = list(additional_list)
        only_nodelist = [x for x in additional_list if isinstance(x, NexusNodeList)]
        self._lists.extend(only_nodelist)



from abc import ABC
from dataclasses import dataclass


@dataclass
class Multir(ABC):
    """Used to represent a single multiplier between regions in a grid."""
    region_1: None | int = None
    region_2: None | int = None
    tmult: None | float = None
    directions: None | str = None
    std_connections: bool = True
    non_std_connections: bool = True



"""Class for collection of Nexus relative permeability and capillary pressure property inputs."""
from __future__ import annotations
from dataclasses import dataclass, field
import os
from typing import Optional, MutableMapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.FcsFile import FcsNexusFile
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusRelPermMethod import NexusRelPermMethod
from ResSimpy.DataModelBaseClasses.RelPerm import RelPerm
from ResSimpy.Utils.dynamic_method_manipulations import add_dynamic_method


@dataclass(kw_only=True)
class NexusRelPermMethods(RelPerm):
    """Class for collection of Nexus relative permeability and capillary pressure property inputs.

    Attributes:
        inputs (dict[int, NexusRelPermMethod]): Collection of Nexus relperm property inputs, as a dictionary
        files (dict[int, NexusFile]): Dictionary collection of relperm property files, as defined in Nexus fcs.
    """

    __inputs: MutableMapping[int, NexusRelPermMethod]
    __files: dict[int, NexusFile]
    __properties_loaded: bool = False  # Used in lazy loading
    __model_unit_system: UnitSystem
    _model_files: Optional[FcsNexusFile] = field(default=None, repr=False, compare=False)

    def __init__(self, model_unit_system: UnitSystem,
                 inputs: Optional[MutableMapping[int, NexusRelPermMethod]] = None,
                 files: Optional[dict[int, NexusFile]] = None, assume_loaded: bool = False,
                 model_files: Optional[FcsNexusFile] = None) -> None:
        """Initialises the NexusRelPermMethods class.

        Args:
            model_unit_system (UnitSystem): Unit system used in the model.
            inputs (Optional[MutableMapping[int, NexusRelPermMethod]]): Collection of Nexus relperm property inputs.
            files (Optional[dict[int, NexusFile]]): Collection of relperm property files, as defined in Nexus fcs file.
                Keyed by the method number.
            assume_loaded (bool): If True, assumes that the properties are already loaded.
            model_files (Optional[FcsNexusFile]): The FcsNexusFile that contains the whole set of model files.
        """
        if inputs:
            self.__inputs = inputs
        else:
            self.__inputs: MutableMapping[int, NexusRelPermMethod] = {}
        if files:
            self.__files = files
        else:
            self.__files = {}
        self.__model_unit_system = model_unit_system
        self.__properties_loaded = assume_loaded
        self._model_files = model_files
        super().__init__()

    def __repr__(self) -> str:
        """Pretty printing relative permeability and capillary pressure methods."""
        if not self.__properties_loaded:
            self.load_relperm_methods()
        printable_str = ''
        for table_num in self.__inputs.keys():
            printable_str += '\n--------------------------------\n'
            printable_str += f'RELPM method {table_num}\n'
            printable_str += '--------------------------------\n'
            printable_str += self.__inputs[table_num].__repr__()
            printable_str += '\n'

        return printable_str

    @property
    def inputs(self) -> MutableMapping[int, NexusRelPermMethod]:
        """Returns collection of Nexus relperm property inputs as dictionary.

        Returns:
            NexusRelPermMethod(MutableMapping): Keyed by method number with values being the NexusRelPermMethod.
        """
        if not self.__properties_loaded:
            self.load_relperm_methods()
        return self.__inputs

    @property
    def files(self) -> dict[int, NexusFile]:
        """Returns dictionary of Nexus Files."""
        return self.__files

    def load_relperm_methods(self) -> None:
        """Loads a collection of relperm Nexus files."""
        # Read in relperm properties from Nexus relperm method files
        if self.__files is not None and len(self.__files) > 0:  # Check if relperm files exist
            for table_num in self.__files.keys():  # For each relperm property method
                relperm_file = self.__files[table_num]
                if relperm_file.location is None:
                    raise ValueError(f'Unable to find relperm file: {relperm_file}')
                if os.path.isfile(relperm_file.location):
                    # Create NexusRelPermMethod object
                    self.__inputs[table_num] = NexusRelPermMethod(file=relperm_file, input_number=table_num,
                                                                  model_unit_system=self.__model_unit_system)
                    # Populate object with relperm properties in file
                    self.__inputs[table_num].read_properties()
        self.__properties_loaded = True

    @property
    def model_unit_system(self) -> UnitSystem:
        """Return the model unit system."""
        return self.__model_unit_system

    @property
    def summary(self) -> str:
        """Returns a string summary of 'NexusRelPermMethods' inputs in a dictionary."""
        relperm_summary = ''
        for key, value in self.inputs.items():
            relperm_summary += f'        {key}: {value.file.location}\n'

        return relperm_summary

    @property
    def keyword(self) -> str:
        """Returns the keyword for the relperm methods."""
        return 'RELPM'

    def add_method(self, method: DynamicProperty, new_file_name: str, create_new_file: bool = False) -> None:
        """Adds a new relpm method to the collection.

        Args:
            method (DynamicProperty): The relpm method to add.
            new_file_name (str): The name of the file to save the method to.
            create_new_file (bool): Whether to create a new file for the method.
        """
        add_dynamic_method(dynamic_method_collection=self, method=method, new_file_name=new_file_name,
                           create_new_file=create_new_file)

    def _method_type(self) -> type[NexusRelPermMethod]:
        """Returns the expected type of the dynamic property."""
        return NexusRelPermMethod



"""Nexus implementation of the Well Connection class."""

from __future__ import annotations
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.Network.NexusStation import NexusStation
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.DataModelBaseClasses.WellConnection import WellConnection


@dataclass(kw_only=True)
class NexusWellConnection(WellConnection):
    """Nexus implementation of the Well Connection class.

    Attributes:
    stream (str): Stream identifier (STREAM)
    number (int): Identification number (NUMBER)
    scale (float): Scaling factor (SCALE)
    bhdepth (float): Bottomhole depth (BHDEPTH)
    datum_depth (float): Depth relative to a datum (DATUM)
    x_pos (float): X-coordinate position (X)
    y_pos (float): Y-coordinate position (Y)
    gradient_calc (str): Gradient calculation method (DATGRAD)
    length (float): Length of the well connection (LENGTH)
    bottomhole_measured_depth (float): Measured depth of the bottom hole (BHMD)
    add_tubing (int): Additional tubing indicator (ADDTUBING)
    diameter (float): Diameter of the well connection (DIAM)
    inner_diameter (float): Inner diameter of the well connection (INNERDIAM)
    roughness (float): Pipe roughness (ROUGHNESS)
    tracer (str): Tracer substance identifier (TRACERS)
    con_type (str): Connection type (TYPE)
    hyd_method (str): Hydraulic method (METHOD)
    pvt_method (int): PVT method (IPVT)
    water_method (int): Water method (IWAT)
    bat_method (int): Bat method (IBAT)
    temperature (float): Temperature (TEMP)
    elevation_profile (str): Elevation profile identifier (ELEVPR)
    temperature_profile (str): Temperature profile identifier (TEMPPR)
    inj_mobility (str): Injection mobility identifier (INJMOB)
    crossshut (str): Cross-shut identifier (CROSS_SHUT)
    crossflow (str): Crossflow identifier (CROSSFLOW)
    on_time (float): On-time duration (ONTIME)
    heat_transfer_coeff (float): Heat transfer coefficient (HTC)
    well_index_mult (float): Well index multiplier (WIMULT)
    productivity_index (float): Productivity index (PI)
    vip_productivity_index (str): VIP productivity index identifier (VIPPI)
    productivity_index_phase (str): Productivity index phase identifier (PIPHASE)
    d_factor (float): Non-Darcy D-factor (D)
    non_darcy_flow_model (str): Non-Darcy flow model identifier (ND)
    non_darcy_flow_method (str): Non-Darcy flow method identifier (DPERF)
    gas_mobility (str): Gas mobility value (GASMOB)
    capillary_number_model (str): Capillary number model identifier (CN)
    dp_add (float): Additional pressure drop (DPADD)
    dt_add (float): Additional temperature change (DTADD)
    rate_mult (float): Rate multiplier (RATEMULT)
    polymer (str): Polymer identifier (POLYMER)
    station (str): Station identifier (STATION)
    station_object (NexusStation): Instance of NexusStation class.
    drill_queue (str): Drill queue identifier (ASSCDR)
    drill_order_benefit (float): Benefit of the drill order (BENEFIT)
    is_activated (bool): Whether the Well Connection has been activated using ACTIVATED / DEACTIVATE. Defaults to True.
    well_index_mult (Optional[float]): Well index multiplier (WIMULT).
    """

    bh_node_name: Optional[str] = None
    wh_node_name: Optional[str] = None

    stream: Optional[str] = None
    number: Optional[int] = None
    scale: Optional[float] = None
    gradient_calc: Optional[str] = None
    bottomhole_measured_depth: Optional[float] = None
    add_tubing: Optional[int] = None
    tracer: Optional[str] = None
    con_type: Optional[str] = None
    pvt_method: Optional[int] = None
    water_method: Optional[int] = None
    bat_method: Optional[int] = None
    elevation_profile: Optional[str] = None
    temperature_profile: Optional[str] = None
    inj_mobility: Optional[str] = None
    on_time: Optional[float] = None
    heat_transfer_coeff: Optional[float] = None
    well_index_mult: Optional[float] = None
    vip_productivity_index: Optional[float] = None
    productivity_index_phase: Optional[str] = None
    d_factor: Optional[float] = None
    non_darcy_flow_model: Optional[str] = None
    non_darcy_flow_method: Optional[str] = None
    gas_mobility: Optional[str] = None
    capillary_number_model: Optional[str] = None
    dp_add: Optional[float] = None
    dt_add: Optional[float] = None
    rate_mult: Optional[float] = None
    polymer: Optional[str] = None
    station: Optional[str] = None
    station_object: Optional[NexusStation] = None
    drill_queue: Optional[str] = None
    drill_order_benefit: Optional[float] = None

    def __init__(self, properties_dict: Optional[DataObjectMixinDictType] = None,
                 date: Optional[str] = None, date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None, name: Optional[str] = None,
                 bh_node_name: Optional[str] = None, wh_node_name: Optional[str] = None,
                 stream: Optional[str] = None, number: Optional[int] = None, scale: Optional[float] = None,
                 gradient_calc: Optional[str] = None, bottomhole_measured_depth: Optional[float] = None,
                 add_tubing: Optional[int] = None, tracer: Optional[str] = None, con_type: Optional[str] = None,
                 pvt_method: Optional[int] = None, water_method: Optional[int] = None, bat_method: Optional[int] = None,
                 elevation_profile: Optional[str] = None, temperature_profile: Optional[str] = None,
                 inj_mobility: Optional[str] = None, crossshut: Optional[str] = None, crossflow: Optional[str] = None,
                 on_time: Optional[float] = None, heat_transfer_coeff: Optional[float] = None,
                 well_index_mult: Optional[float] = None, vip_productivity_index: Optional[float] = None,
                 productivity_index_phase: Optional[str] = None, d_factor: Optional[float] = None,
                 non_darcy_flow_model: Optional[str] = None, non_darcy_flow_method: Optional[str] = None,
                 gas_mobility: Optional[str] = None, capillary_number_model: Optional[str] = None,
                 dp_add: Optional[float] = None, dt_add: Optional[float] = None, rate_mult: Optional[float] = None,
                 polymer: Optional[str] = None, station: Optional[str] = None,
                 station_object: Optional[NexusStation] = None, drill_queue: Optional[str] = None,
                 drill_order_benefit: Optional[float] = None, bhdepth: Optional[float] = None,
                 datum_depth: Optional[float] = None, x_pos: Optional[float] = None,
                 y_pos: Optional[float] = None, length: Optional[float] = None, temperature: Optional[float] = None,
                 diameter: Optional[float] = None, roughness: Optional[float] = None,
                 inner_diameter: Optional[float] = None, productivity_index: Optional[float] = None,
                 hyd_method: Optional[str] = None, group: Optional[str] = None, i: Optional[int] = None,
                 j: Optional[int] = None, drainage_radius: Optional[float] = None) -> None:
        """Initialises the NexusWellConnection class.

        Args:
            properties_dict (Optional[dict]): dict of the properties to set on the object.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            is_activated: bool: Whether the Well Connection has been set as activated. Defaults to True.
            name (Optional[str]): The name of the object.
            bh_node_name (Optional[str]): Bottom hole node name.
            wh_node_name (Optional[str]): Wellhead node name.
            stream (Optional[str]): Stream identifier (STREAM).
            number (Optional[int]): Identificaiton number (Number).
            scale (Optional[float]): Scaling factor (SCALE).
            gradient_calc (Optional[str]): Gradient calculation method (DATGRAD).
            bottomhole_measured_depth (Optional[float]): Measured depth of the bottom hole (BHMD).
            add_tubing (Optional[int]): Additional tubing indicator (ADDTUBING).
            tracer (Optional[str]): Tracer substance identifier (TRACERS).
            con_type (Optional[str]): Connection type (TYPE).
            pvt_method (Optional[int]): PVT method (IPVT).
            water_method (Optional[int]): Water method (IWAT).
            bat_method (Optional[int]): Bat method (IBAT).
            elevation_profile (Optional[str]): Elevation profile identifier (ELEVPR).
            temperature_profile (Optional[str]): Temperature profile identifier (TEMPPR).
            inj_mobility (Optional[str]): Injection mobility identifier (INJMOB).
            crossshut (Optional[str]): Cross-shut identifier (CROSS_SHUT).
            crossflow (Optional[str]): Crossflow identifier (CROSSFLOW).
            on_time (Optional[float]): On-time duration (ONTIME).
            heat_transfer_coeff (Optional[float]): Heat transfer coefficient (HTC).
            vip_productivity_index (Optional[float]): VIP productivity index identifier (VIPPI).
            productivity_index_phase (Optional[str]): Productivity index phase identifier (PIPHASE).
            d_factor (Optional[float]): Non-Darcy D-factor (D).
            non_darcy_flow_model (Optional[str]): Non-Darcy flow model identifier (ND).
            non_darcy_flow_method (Optional[str]): Non-Darcy flow method identifier (DPERF)
            gas_mobility (Optional[float]): Gas mobility value (GASMOB).
            capillary_number_model (Optional[str]): Capillary number model identifier (CN).
            dp_add (Optional[float]): Additional pressure drop (DPADD).
            dt_add (Optional[float]): Additional temperature change (DTADD).
            rate_mult (Optional[float]): Rate multiplier (RATEMULT).
            polymer (Optional[str]): Polymer identifier (POLYMER).
            station (Optional[str]): Station identifier (STATION).
            station_object (NexusStation): Instance of NexusStation class.
            drill_queue (Optional[str]): Drill queue identifier (ASSCDR).
            drill_order_benefit (Optional[float]): Benefit of the drill order (BENEFIT).
            is_activated (bool): Whether the Well Connection has been activated using ACTIVATED / DEACTIVATE
            Defaults to True.
            bhdepth (float): Bottomhole depth (BHDEPTH)
            datum_depth (float): Depth relative to a datum (DATUM)
            x_pos (float): X-coordinate position (X)
            y_pos (float): Y-coordinate position (Y)
            length (float): Length of the well connection (LENGTH)
            temperature (float): Temperature (TEMP)
            diameter (float): Diameter of the well connection (DIAM)
            roughness (float): Pipe roughness (ROUGHNESS)
            inner_diameter (float): Inner diameter of the well connection (INNERDIAM)
            productivity_index_phase (str): Productivity index phase identifier (PIPHASE)
            hyd_method (str): Hydraulic method (METHOD)
            group (Optional[str]): The group that the Well Connection belongs to.
            drainage_radius (Optional[float]): The drainage radius.
            i (Optional[int]): The location of the Well Connection in the i direction.
            j (Optional[int]): The location of the Well Connection in the j direction.
            productivity_index (Optional[float]): Productivity index (PI).
            well_index_mult (Optional[float]): Well index multiplier (WIMULT).
        """

        self.bh_node_name = bh_node_name
        self.wh_node_name = wh_node_name
        self.stream = stream
        self.number = number
        self.scale = scale
        self.gradient_calc = gradient_calc
        self.bottomhole_measured_depth = bottomhole_measured_depth
        self.add_tubing = add_tubing
        self.tracer = tracer
        self.con_type = con_type
        self.pvt_method = pvt_method
        self.water_method = water_method
        self.bat_method = bat_method
        self.elevation_profile = elevation_profile
        self.temperature_profile = temperature_profile
        self.inj_mobility = inj_mobility
        self.on_time = on_time
        self.heat_transfer_coeff = heat_transfer_coeff
        self.well_index_mult = well_index_mult
        self.vip_productivity_index = vip_productivity_index
        self.productivity_index_phase = productivity_index_phase
        self.d_factor = d_factor
        self.non_darcy_flow_model = non_darcy_flow_model
        self.non_darcy_flow_method = non_darcy_flow_method
        self.gas_mobility = gas_mobility
        self.capillary_number_model = capillary_number_model
        self.dp_add = dp_add
        self.dt_add = dt_add
        self.rate_mult = rate_mult
        self.polymer = polymer
        self.station = station
        self.station_object = station_object
        self.drill_queue = drill_queue
        self.drill_order_benefit = drill_order_benefit
        super().__init__(date_format=date_format, start_date=start_date, unit_system=unit_system, name=name, date=date,
                         properties_dict=properties_dict, bhdepth=bhdepth, datum_depth=datum_depth, x_pos=x_pos,
                         y_pos=y_pos, length=length, temperature=temperature, diameter=diameter, roughness=roughness,
                         inner_diameter=inner_diameter, productivity_index=productivity_index, hyd_method=hyd_method,
                         group=group, i=i, j=j, drainage_radius=drainage_radius, well_index_mult=well_index_mult,
                         crossflow=crossflow, crossshut=crossshut)

        if self.name is not None:
            self.bh_node_name = self.name + '%bh'
            self.wh_node_name = self.name + '%wh'

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        nexus_mapping = {
            'NAME': ('name', str),
            'STREAM': ('stream', str),
            'NUMBER': ('number', int),
            'SCALE': ('scale', float),
            'BHDEPTH': ('bhdepth', float),
            'DATUM': ('datum_depth', float),
            'X': ('x_pos', float),
            'Y': ('y_pos', float),
            'DATGRAD': ('gradient_calc', str),
            'LENGTH': ('length', float),
            'BHMD': ('bottomhole_measured_depth', float),
            'ADDTUBING': ('add_tubing', int),
            'DIAM': ('diameter', float),
            'INNERDIAM': ('inner_diameter', float),
            'ROUGHNESS': ('roughness', float),
            'TRACERS': ('tracer', str),
            'TYPE': ('con_type', str),
            'METHOD': ('hyd_method', str),
            'IPVT': ('pvt_method', int),
            'IWAT': ('water_method', int),
            'IBAT': ('bat_method', int),
            'TEMP': ('temperature', float),
            'ELEVPR': ('elevation_profile', str),
            'TEMPPR': ('temperature_profile', str),
            'INJMOB': ('inj_mobility', str),
            'CROSS_SHUT': ('crossshut', str),
            'CROSSFLOW': ('crossflow', str),
            'ONTIME': ('on_time', float),
            'HTC': ('heat_transfer_coeff', float),
            'WIMULT': ('well_index_mult', float),
            'PI': ('productivity_index', float),
            'VIPPI': ('vip_productivity_index', float),
            'PIPHASE': ('productivity_index_phase', str),
            'D': ('d_factor', float),
            'ND': ('non_darcy_flow_model', str),
            'DPERF': ('non_darcy_flow_method', str),
            'GASMOB': ('gas_mobility', str),
            'CN': ('capillary_number_model', str),
            'DPADD': ('dp_add', float),
            'DTADD': ('dt_add', float),
            'RATEMULT': ('rate_mult', float),
            'POLYMER': ('polymer', str),
            'STATION': ('station', str),
            'ASSCDR': ('drill_queue', str),
            'BENEFIT': ('drill_order_benefit', float),
            }
        return nexus_mapping

    def __repr__(self) -> str:
        return super().__repr__()
